/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file c1rect2.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _C1RECT2_HPP
#define _C1RECT2_HPP

#include <sstream>
#include "capd/minmax.h"
#include "dynset/c1rect2intf.h"

namespace capd{
namespace dynset{

template<typename MatrixType>
C1rect2<MatrixType>::C1rect2(const NormType& aNorm, int dim)
  : m_x(dim), m_r(dim), m_r0(dim),
    m_B(MatrixType::Identity(dim)),
    m_C(MatrixType::Identity(dim)),
    m_D(MatrixType::Identity(dim)),
    m_Bjac(MatrixType::Identity(dim)),
    m_R(dim,dim),
    m_Cjac(MatrixType::Identity(dim)),
    m_R0(dim,dim),
    m_Norm(aNorm.clone()),
    m_sizeFactor(20.)
{
  m_x.clear();
  m_r.clear();
  m_r0.clear();
  m_R.clear();
  m_R0.clear();
}


template<typename MatrixType>
C1rect2<MatrixType>::C1rect2(const C1rect2& s)
  : m_x(s.m_x),
    m_r(s.m_r),
    m_r0(s.m_r0),
    m_B(s.m_B),
    m_C(s.m_C),
    m_D(s.m_D),
    m_Bjac(s.m_Bjac),
    m_R(s.m_R),
    m_Cjac(s.m_Cjac),
    m_R0(s.m_R0),
    m_Norm(s.m_Norm->clone()),
    m_sizeFactor(s.m_sizeFactor)
{}

template<typename MatrixType>
C1rect2<MatrixType>::C1rect2(const VectorType& x, const NormType& aNorm, double sizeFactor)
 : m_x(x),
   m_r(x.dimension()),
   m_r0(x.dimension()),
   m_B(MatrixType::Identity(x.dimension())),
   m_C(MatrixType::Identity(x.dimension())),
   m_D(MatrixType::Identity(x.dimension())),
   m_Bjac(MatrixType::Identity(x.dimension())),
   m_R(x.dimension(),x.dimension()),
   m_Cjac(MatrixType::Identity(x.dimension())),
   m_R0(x.dimension(),x.dimension()),
   m_Norm(aNorm.clone()),
   m_sizeFactor(sizeFactor)
{
  split(m_x,m_r0);
  m_r.clear();
  m_R.clear();
  m_R0.clear();
}

template<typename MatrixType>
C1rect2<MatrixType>::C1rect2(
  const VectorType& x,
  const VectorType& r0,
  const NormType& aNorm,
  double sizeFactor
  ):m_x(x),
    m_r(x.dimension()),
    m_r0(r0),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(MatrixType::Identity(x.dimension())),
    m_D(MatrixType::Identity(x.dimension())),
    m_Bjac(MatrixType::Identity(x.dimension())),
    m_R(x.dimension(),x.dimension()),
    m_Cjac(MatrixType::Identity(x.dimension())),
    m_R0(x.dimension(),x.dimension()),
    m_Norm(aNorm.clone()),
    m_sizeFactor(sizeFactor)
{
  m_x += m_r0;
  split(m_x,m_r0);
  m_r.clear();
  m_R.clear();
  m_R0.clear();  
}


template<typename MatrixType>
C1rect2<MatrixType>::C1rect2(
  const VectorType& x,
  const MatrixType& C,
  const VectorType& r0,
  const NormType& aNorm,
  double sizeFactor
  ):m_x(x),
    m_r(x.dimension()),
    m_r0(r0),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(C),
    m_D(MatrixType::Identity(x.dimension())),
    m_Bjac(MatrixType::Identity(x.dimension())),
    m_R(x.dimension(),x.dimension()),
    m_Cjac(MatrixType::Identity(x.dimension())),
    m_R0(x.dimension(),x.dimension()),
    m_Norm(aNorm.clone()),
    m_sizeFactor(sizeFactor)
{
  m_r.clear();
  m_R.clear();
  m_R0.clear();
// we must assure that m_x\in m_x + m_C*m_r0
  if(!subset(m_r,m_r0))
  {
    VectorType centerR0=m_r0;
    split(centerR0,m_r0);
    // now the m_r0 is centered. We must add C*centerR0 to m_x
    m_x += m_C*centerR0;
    // finally, we split m_x to the point ane the rest is added to r
    split(m_x,m_r);
  }
}

//-----------------------------------------------

template<typename MatrixType>
void C1rect2<MatrixType>::move(capd::dynsys::C1DynSys<MatrixType> &c1dynsys, C1rect2 &result) const
{
  int dim = m_x.dimension();
  VectorType Rem(dim);
  MatrixType jacRem(dim,dim);

  VectorType xx = m_x + m_C*m_r0 + m_B*m_r;
  
  int found;
  VectorType Enc = c1dynsys.enclosure(xx,&found);
  MatrixType S = c1dynsys.jacEnclosure(Enc,*m_Norm);
  c1dynsys.JacRemainder(Enc,S,Rem,jacRem);
  MatrixType Jac = c1dynsys.JacPhi(xx);

  result.m_x = c1dynsys.Phi(m_x) + Rem;
// xx is unnecesarry now
  split(result.m_x,xx);

  result.m_C = Jac*m_C;
  split(result.m_C,S);
  xx += S*m_r0;
  MatrixType A = Jac*m_B;
// S is unnecessary now
  mid(A,result.m_B); // result.m_B = midMatrix(A) but faster
  orthonormalize(result.m_B);
  MatrixType Qtr = Transpose(result.m_B);
  result.m_r = (Qtr*A)*m_r + Qtr*xx;

  if(&result!=this) result.m_r0 = m_r0;
  // T.Kapela - put set r into r0 if r is too big.
  if(m_sizeFactor!=0)
    if( ::size(result.m_r) > m_sizeFactor * ::size(result.m_r0))
    {
      result.m_r0 = result.m_r +  ((Qtr * result.m_C) * result.m_r0);
      result.m_C = result.m_B;
      result.m_B = MatrixType::Identity(dim);
      result.m_r = ScalarType(0.);
    }

//-----------------------------------------------

  Jac += jacRem;
  split(Jac,S);
  result.m_D = Jac*m_D + S*(m_D + m_Bjac*m_R + m_Cjac*m_R0);
  split(result.m_D,A);

  result.m_Cjac = Jac * m_Cjac;
  split(result.m_Cjac,S);

  A += S*m_R0;
  // S is unnecessary now, so we will use it again
  S = Jac*m_Bjac;
  mid(S,result.m_Bjac);
  orthonormalize(result.m_Bjac);
  Qtr = Transpose(result.m_Bjac);
  result.m_R = (Qtr*S)*m_R + Qtr*A;

  if (&result!=this) 
    result.m_R0 = m_R0;

  if(m_sizeFactor!=0)
  {
    typename ScalarType::BoundType maxSizeR = 0., maxSizeR0 = 0.;
    int i,j;
    for(i=0;i<dim;i++)
      for(j=0;j<dim;j++)
      {
        maxSizeR = capd::max(maxSizeR,diam(result.m_R[i][j]).rightBound());
        maxSizeR0 = capd::max(maxSizeR0, diam(result.m_R0[i][j]).rightBound());
      }

    if(maxSizeR>m_sizeFactor*maxSizeR0)
    {
      result.m_R0 = result.m_R + (Qtr * result.m_Cjac) * result.m_R0;
      result.m_Cjac = result.m_Bjac;
      result.m_Bjac = MatrixType::Identity(dim);
      result.m_R.clear();
    }
  }
}

// ------------------------------------------------------------------------

template<typename MatrixType>
typename C1rect2<MatrixType>::ScalarType C1rect2<MatrixType>::size(void) const
{
  return ::size(m_x + m_C*m_r0 + m_B*m_r);
}

template<typename MatrixType>
std::string C1rect2<MatrixType>::show(void) const
{
  std::ostringstream descriptor;
  descriptor << "C1rect2:\n x=";
  descriptor << m_x << "\n B=";
  descriptor << m_B << "\n r=";
  descriptor << m_r << "\n C= ";
  descriptor << m_C << "\n r0= ";
  descriptor << m_r0 << "\n D= ";
  descriptor << m_D << "\n Bjac= ";
  descriptor << m_Bjac << "\n R= ";
  descriptor << m_R << "\n Cjac= ";
  descriptor << m_Cjac << "\n R0= ";
  descriptor << m_R0 << '\0';
  
  return descriptor.str();
}

template<typename MatrixT>
C1rect2<MatrixT>::operator typename C1rect2<MatrixT>::MatrixType() const
{
  return m_D + m_Bjac*m_R + m_Cjac*m_R0;
}

template<typename MatrixType>
C1rect2<MatrixType> &C1rect2<MatrixType>::operator=(const VectorType &x)
{
  int dim = x.dimension();
  m_x = x;
  m_r0 = VectorType(dim);
  split(m_x,m_r0);
  m_C = MatrixType::Identity(dim);
  m_B = MatrixType::Identity(dim);
  m_r = VectorType(dim);
  m_r.clear();

  m_Bjac = m_Cjac = m_D = MatrixType::Identity(dim);
  m_R = MatrixType(dim,dim);
  m_R.clear();
  m_R0 = MatrixType(dim,dim);
  m_R0.clear();
  
  return *this;
}

template<typename MatrixType>
C1rect2<MatrixType>& C1rect2<MatrixType>::operator=(const C1rect2 &S)
{
  m_x = S.m_x;
  m_B = S.m_B;
  m_r = S.m_r;
  m_C = S.m_C;
  m_r0 = S.m_r0;
  
  m_D = S.m_D;
  m_Bjac = S.m_Bjac;
  m_R = S.m_R;
  m_Cjac = S.m_Cjac;
  m_R0 = S.m_R0;
  
  m_sizeFactor = S.m_sizeFactor;
  delete m_Norm;
  m_Norm = S.m_Norm->clone();
  return *this;
}

}} // namespace capd::dynset

#endif // _C1RECT_HPP

/// @}
