/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file c1rect2intf.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2002-2004 */

#ifndef _C1rect2Intf_H
#define _C1rect2Intf_H

#include "dynset/c1set.h"
#include "vectalg/normintf.h"
#include "dynsys/c1dynsys.h"

/* C^1-Lohner algorithm.
   Derivative of the flow moved via QR decomposition (third method)
   the set part - QR decomposition (3-rd method)
*/

namespace capd{
namespace dynset{

template<typename MatrixT>
class C1rect2 : public C1set<MatrixT>{
public:
   typedef MatrixT MatrixType;
   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;
   typedef Norm<VectorType,MatrixType> NormType;

   C1rect2(const NormType& the_norm, int);
   C1rect2(const VectorType& the_x, const NormType&, double s_factor = 20);
   C1rect2(const VectorType& the_x, const VectorType& the_r, const NormType&, double s_factor = 20);
   C1rect2(const VectorType& the_x, const MatrixType&, const VectorType& the_r, const NormType&, double s_factor = 20);
   C1rect2(const C1rect2 &c);
   ~C1rect2(){delete m_Norm;}

   ScalarType size(void) const;
   C1set<MatrixType>* clone(void) const;
   C1set<MatrixType>* fresh(void) const;
   C1set<MatrixType>* cover(const VectorType& v) const;

   VectorType center(void) const;
   void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys);
   void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys, C1rect2& result) const;

   std::string show(void) const;
   NormType* getNorm(void);
   double setFactor(double sFactor); // sets size factor :  0 = no swapping
   void disableSwapping();
   double getFactor();

   operator VectorType() const;
   operator MatrixType() const;
   C1rect2 &operator=(const VectorType &v);
   C1rect2 &operator=(const C1rect2 &S);

protected:

// a representation of the set: x + C*r0 + B*r

   VectorType m_x,m_r,m_r0;
   MatrixType m_B,m_C;

   MatrixType m_D, m_Bjac, m_R, m_Cjac, m_R0; // D + Bjac*R  + Cjac * R0;

   NormType *m_Norm; //the logarithmic Norm<IVectorType,IMatrixType> used in the computation of a rough
                    // enclosure for the variational part

   double m_sizeFactor;  // if size of r0 is greater than r *size_factor we  put r in r0 and set r=0
                       // if size_factor = 0 we disable swapping
};

// -------------------------------------------------------------------

template<typename MatrixType>
inline void C1rect2<MatrixType>::move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys)
{
   move(c1dynsys,*this);
}

template<typename MatrixType>
inline C1set<MatrixType>* C1rect2<MatrixType>::clone(void) const
{
     return new C1rect2<MatrixType>(*this);
}

template<typename MatrixType>
inline C1set<MatrixType>* C1rect2<MatrixType>::fresh(void) const
{
   return new C1rect2<MatrixType>(*m_Norm,m_x.dimension());
}

template<typename MatrixType>
inline C1set<MatrixType>* C1rect2<MatrixType>::cover(const VectorType& v) const
{
   return new C1rect2<MatrixType>(v, *m_Norm);
}

template<typename MatrixType>
inline typename C1rect2<MatrixType>::VectorType C1rect2<MatrixType>::center(void) const
{
   return m_x;
}

template<typename MatrixType>
inline C1rect2<MatrixType>::operator typename C1rect2<MatrixType>::VectorType(void) const
{
   return m_x + m_C*m_r0 + m_B*m_r;
}

template<typename MatrixType>
inline typename C1rect2<MatrixType>::NormType* C1rect2<MatrixType>::getNorm(void)
{
   return m_Norm;
}

template<typename MatrixType>
inline double C1rect2<MatrixType>::setFactor(double sFactor) // sets size factor :  0 = no swapping
{
   double s = m_sizeFactor;
   m_sizeFactor = sFactor;
   return s;
}

template<typename MatrixType>
inline void C1rect2<MatrixType>::disableSwapping()
{
   m_sizeFactor = 0.0;
}

template<typename MatrixType>
inline double C1rect2<MatrixType>::getFactor() // returns size factor
{
   return m_sizeFactor;
}

}} // the end of the namespace capd::dynset

#endif // C1rect2Intf_H

/// @}
