/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cnrect2.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2006 */

#ifndef _CnRect2_HPP
#define _CnRect2_HPP

#include "dynset/cnrect2intf.h"

namespace capd{
namespace dynset{

// ----------------------------- CONSTRUCTORS --------------------------------

template<typename MatrixT>
CnRect2<MatrixT>::CnRect2(const VectorType& x,  const NormType& a_norm, int rank, double sizeFactor)
  : m_x(x.dimension(),rank),
    m_r(x.dimension(),rank),
    m_r0(x.dimension(),rank),
    m_currentSet(x.dimension(),rank),
    m_B(MatrixT::Identity(x.dimension())),
    m_C(MatrixT::Identity(x.dimension())),
    m_Bjac(MatrixT::Identity(x.dimension())),
    m_Cjac(MatrixT::Identity(x.dimension())),
    m_sizeFactor(sizeFactor),
    m_norm(a_norm.clone())
{
  m_currentSet() = x;
  m_x() = x;
  typename MatrixT::RefColumnVectorType refX = m_x();
  typename MatrixT::RefColumnVectorType refR0 = m_r0();
  split(refX,refR0);

// initial condition for variational part is an Identity matrix
  if(getRank()>0)
  {
    for(int i=0;i<getDimension();++i)
      m_currentSet(i,i) = m_x(i,i) = ScalarType(1);
  }
}

// ----------------------------------------------------------------------------

template<typename MatrixT>
CnRect2<MatrixT>::CnRect2(const VectorType& x, const MatrixType& C, const VectorType& r0, const NormType& a_norm, int rank, double sizeFactor)
  : m_x(x.dimension(),rank),
    m_r(x.dimension(),rank),
    m_r0(x.dimension(),rank),
    m_currentSet(x.dimension(),rank),
    m_B(MatrixT::Identity(x.dimension())),
    m_C(C),
    m_Bjac(MatrixT::Identity(x.dimension())),
    m_Cjac(MatrixT::Identity(x.dimension())),
    m_sizeFactor(sizeFactor),
    m_norm(a_norm.clone())
{
  m_x() = x;
  m_r0() = r0;
  VectorType v = x + C*r0;
  m_currentSet() = v;

  typename MatrixT::RefColumnVectorType refX = m_x();
  typename MatrixT::RefColumnVectorType refR0 = m_r0();
  typename MatrixT::RefColumnVectorType refR = m_r();
  if(!subset(refR,refR0))
  {
    VectorType copyX = x;
    VectorType copyR0 = r0;
    VectorType centerR0(x.dimension());
    split(centerR0,copyR0);
    copyX += m_C*centerR0;
    m_r0() = copyR0;
    split(copyX,copyR0);
    m_r() = copyR0;
  }

// initial condition for variational part is an Identity matrix
  if(getRank()>0)
  {
    for(int i=0;i<getDimension();++i)
      m_currentSet(i,i) = m_x(i,i) = ScalarType(1);
  }
}

// ----------------------------------------------------------------------------

template<typename MatrixT>
CnRect2<MatrixT>::CnRect2(const CnRect2& a_set)
  : m_x(a_set.m_x),
    m_r(a_set.m_r),
    m_r0(a_set.m_r0),
    m_currentSet(a_set.m_currentSet),
    m_B(a_set.m_B),
    m_C(a_set.m_C),
    m_Bjac(a_set.m_Bjac),
    m_Cjac(a_set.m_Cjac),
    m_sizeFactor(a_set.m_sizeFactor),
    m_norm(a_set.m_norm->clone())
{}

// ---------------------------- operators -------------------------------------

template<typename MatrixT>
CnRect2<MatrixT>& CnRect2<MatrixT>::operator=(const CnRect2& s)
{
  if(this!=&s) 
  {
    m_x = s.m_x;
    m_r = s.m_r;
    m_r0 = s.m_r0;
    m_currentSet = s.m_currentSet;
    m_B = s.m_B;
    m_C = s.m_C;
    m_Bjac = s.m_Bjac;
    m_Cjac = s.m_Cjac;
    m_sizeFactor = s.m_sizeFactor;
    delete m_norm;
    m_norm = (s.m_norm)->clone();
  }
  return *this;
}

// ----------------------------------------------------------------------------

template<typename MatrixT>
CnRect2<MatrixT>& CnRect2<MatrixT>::operator=(const VectorType& v)
{
  if(v.dimension()!=getDimension())
    throw std::runtime_error("CnRect2::operator=(Vector) - incorrect dimensions");
  VectorType temp(getDimension());
  VectorType y = v;
  split(y,temp);
  m_x.clear();
  m_r.clear();
  m_r0.clear();
  m_currentSet.clear();
  m_B = m_C = m_Bjac = m_Cjac = MatrixType::Identity(getDimension());
   
  m_x() = y;
  m_r0() = temp;
  m_currentSet() = v;
  if(getRank()>0)
  {
    for(int i=0;i<getDimension();++i)
      m_currentSet(i,i) = ScalarType(1);
  }
  return *this;
}

// ------------------------------ move ---------------------------------------

template<typename MatrixT>
void CnRect2<MatrixT>::move(capd::dynsys::CnDynSys<MatrixType>& cndynsys, CnRect2& result) const
{
  int dimension = getDimension();
  VectorType xx(*this);
  VectorType x(m_x), s(dimension);
  MatrixType S(dimension,dimension);
  CnCoeffType Phi(dimension,getRank()),
              Rem(dimension,getRank());
   
  cndynsys.cnPhi(xx,Phi);
  // try to correct DPhi by using of second order derivatives and the mean value theorem
  if(getRank()>1)
  {
    MatrixType M = cndynsys.JacPhi(x);
    VectorType deltaX = xx-x;
    for(int i=0;i<dimension;++i)
      for(int j=0;j<dimension;++j)
      {
        ScalarType ref = M(i+1,j+1);
        for(int k=0;k<dimension;++k)
          ref += Phi(i,j,k)*deltaX[k];
        if(!intersection(Phi(i,j),ref,Phi(i,j)))
        {
          std::cout << x << std::endl;
          std::cout << xx << std::endl;
          std::cout << m_r() << std::endl;
          std::cout << m_r0() << std::endl;
          throw std::runtime_error("CnRect2::move - empty intersection when computing D\\phi");
        }
      }
  }
  cndynsys.cnRemainder(xx,*m_norm,Rem);

// C^0 part
  VectorType y = cndynsys.Phi(x);
  y += VectorType(Rem());
  split(y,s);
  MatrixType A = MatrixType(Phi);
  result.m_C = A*m_C;
  split(result.m_C,S);
  s += S * m_r0();
  MatrixType D = A*m_B;
  MatrixType Q = midMatrix(D);
  orthonormalize(Q);
  MatrixType Qtr = Transpose(Q);
  result.m_x() = y;
  result.m_r() = (Qtr*D)*m_r() + Qtr*s;
  result.m_B = Q;
  if(this!=&result)
    result.m_r0() = m_r0();

  if(m_sizeFactor!=0)
    if( ::size(result.m_r()) > m_sizeFactor * ::size(result.m_r0()))
    {
      result.m_r0() = result.m_r() +  ((Qtr * result.m_C) * result.m_r0());
      result.m_C = result.m_B;
      result.m_B = MatrixType::Identity(dimension);
      result.m_r().clear();
    }

// C^1 part
//   Phi += Rem

  typename CnCoeffType::iterator b=Phi.begin(), e=Phi.end(), j=Rem.begin();
  while(b!=e)
  {
    (*b) += (*j);
    ++b;
    ++j;
  }

  A = MatrixType(Phi); // in fact A += Rem
  MatrixType deltaA(dimension,dimension);
  split(A,deltaA);
  result.m_Cjac = A * m_Cjac;
  split(result.m_Cjac,S);
   
  D = A*MatrixType(m_x) + deltaA*MatrixType(m_currentSet);
  MatrixType Z(dimension,dimension);
  Z += S*MatrixType(m_r0);
  split(D,Z);
  MatrixType X = A*m_Bjac;
  Q = midMatrix(X);
  orthonormalize(Q);
  Qtr = Transpose(Q);
  X=Qtr*X;
  MatrixType R = X*MatrixType(m_r) + Qtr*Z;
  result.m_r.setMatrix(R);
  result.m_Bjac = Q;
  result.m_x.setMatrix(D);
      
  MatrixType R0 = MatrixType(m_r0);
  if(&result!=this) 
    result.m_r0.setMatrix(R0);

// C^2 -- C^r part
  // Rem is now unnecessary. We will store in Rem a nonlinear part of composition
  if(getRank()>1)
  {
    substitutionPowerSeries(Phi,m_currentSet,Rem,true);
  }
  result.m_currentSet() = y + result.m_C*result.m_r0() + result.m_B*result.m_r();
  result.m_currentSet.setMatrix(D + result.m_Cjac * R0 + result.m_Bjac * R);

  for(int p=2;p<=getRank();++p)
  {
    Multipointer mp = m_x.first(p);
    do{
      y = Rem(mp) + A*m_x(mp) + deltaA*m_currentSet(mp);
      s += S*m_r0(mp);
      split(y,s);
      VectorType r = (Qtr*s) + (X*m_r(mp));
      result.m_r(mp) = r;
      result.m_x(mp) = y;
      if(&result!=this)
        result.m_r0(mp) = m_r0(mp);
      result.m_currentSet(mp) = y + Q*r + result.m_Cjac * result.m_r0(mp);
    }while(m_x.next(mp));
  }

  if(m_sizeFactor!=0)
  {
    typename ScalarType::BoundType maxSizeR = 0., maxSizeR0 = 0.;
    for(int i=0;i<dimension;i++)
      for(int j=0;j<dimension;j++)
      {
        maxSizeR = capd::max(maxSizeR,diam(R[i][j]).rightBound());
        maxSizeR0 = capd::max(maxSizeR0, diam(R0[i][j]).rightBound());
      }

      if(maxSizeR>m_sizeFactor*maxSizeR0)
      {
        for(int p=1;p<=getRank();++p)
        {
          Multipointer mp = m_x.first(p);
          do{
            result.m_r0(mp) = result.m_r(mp) + (Qtr * result.m_Cjac) * result.m_r0(mp);
            result.m_r(mp).clear();
          }while(m_x.next(mp));
        }
        result.m_Cjac = result.m_Bjac;
        result.m_Bjac = MatrixType::Identity(dimension);
      }
  }
}

}} // namespace capd::dynset

#endif // _CnRect2_HPP

/// @}
