/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file basicCnTaylornIntf.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2006 */

#ifndef _BasicCnTaylorIntf_H
#define _BasicCnTaylorIntf_H
#include <utility>
#include <algorithm>

namespace capd{
namespace dynsys{

template<typename CnMapT>
class BasicCnTaylor{
public:
   typedef CnMapT CnMapType;
   typedef CnMapT MapType;
   typedef typename CnMapT::FunctionType FunctionType;
   typedef typename CnMapType::MatrixType MatrixType;
   typedef typename MatrixType::RowVectorType VectorType;
   typedef typename MatrixType::ScalarType ScalarType;
   typedef typename CnMapType::C2CoeffType C2CoeffType;
   typedef typename CnMapType::CnCoeffType CnCoeffType;
   typedef typename CnMapType::NodeType NodeType;
   typedef typename CnMapType::ConsNodeType ConsNodeType;
   typedef typename CnMapType::SeriesContainerType SeriesContainerType;
   typedef typename SeriesContainerType::Multipointer Multipointer;
   typedef typename SeriesContainerType::Multiindex Multiindex;

   BasicCnTaylor(const CnMapType& a_vectorField, int a_order, const ScalarType& a_step); // rank will be read form vectorField object
// for compatibility with classes Taylor and C2Taylor
   BasicCnTaylor(const std::string& a_vectorField, int a_order, const ScalarType& a_step);
   BasicCnTaylor(const char* a_vectorField, int a_order, const ScalarType& a_step);
   virtual ~BasicCnTaylor();

   VectorType operator()(const VectorType& v) const;
   VectorType operator()(const VectorType& v, MatrixType& der) const;
   VectorType operator()(CnCoeffType&) const;

   void setField(const CnMapType& newVectorField){} //TODO
   void setOrder(int newOrder);

// inline definitions
   inline void setStep(const ScalarType& a_step) { m_step = a_step; }
   inline const CnMapType& getField() const { return m_vectorField; }
   inline int getOrder() const { return m_order; }
   inline int getRank() const { return m_vectorField.getRank(); }
   inline const ScalarType& getStep() const { return m_step; }
   inline int getDimension() const { return m_vectorField.getDimension(); } 
   inline void setParameter(const std::string&p, const ScalarType&v)  { m_vectorField.setParameter(p,v); }

   const SeriesContainerType& computeDaPhi(const CnCoeffType& c, int r) const;

protected:
   CnMapType m_vectorField;
   ScalarType m_step;
   int m_order;
   SeriesContainerType m_vectorFieldSeries;
   SeriesContainerType m_resultSeries;
   SeriesContainerType m_compositionFormula;
   SeriesContainerType m_nonlinearPart;
   ConsNodeType m_zeroNode;
   
   typename Multiindex::IndicesSet m_listIndices;

   // this is the main function which computes a composition formula 
   void computeCompositionFormula();
   void computeCompositionFormula(const Multipointer&);
   NodeType& computeProduct(const Multiindex&, const Multipointer&, int p, int k);

   void operator=(const BasicCnTaylor&){}
   BasicCnTaylor(const BasicCnTaylor&): m_vectorFieldSeries(1,1,1), 
                                        m_resultSeries(1,1,1), 
                                        m_compositionFormula(1,1,1), 
                                        m_nonlinearPart(1,1,1), 
                                        m_zeroNode(ScalarType(0),1)
                                        {}
   void setInitialCondition(const CnCoeffType& coeff) const;
   void setInitialCondition(const VectorType& v, const MatrixType& m, const C2CoeffType& hessian) const;
   void setInitialCondition(const VectorType& v, const MatrixType& m) const;
   void setInitialCondition(const VectorType& v) const;
   void computeVectorFieldSeries(const VectorType& v, int rank, int order) const;
   void computeResultSeries(int rank, int a_order) const;
   void computeNonlinearPart(int rank) const;
}; // the end of class BasicCnTaylor

}} // the end of the namespace capd::dynsys

#endif // _BasicCnTaylorIntf_H

/// @}
