/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file taylor.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2001-2004 */

#ifndef _TAYLOR_HPP
#define _TAYLOR_HPP

#include <sstream>
#include <string>
#include <stdexcept>

#include "capd/factrial.h"
#include "dynsys/taylorex.h"
#include "dynsys/taylorintf.h"
#include "dynsys/enclosure.h"

namespace capd{
namespace dynsys{

//###########################################################//

template<typename MapType>
Taylor<MapType>::Taylor(const MapType& vectorField, int a_order, const ScalarType& Step)
   : BasicTaylor<MapType>(vectorField,a_order,Step)
{}

//###########################################################//

template<typename MapType>
Taylor<MapType>::Taylor(const std::string &vectorField, int a_order, const ScalarType &Step)
   : BasicTaylor<MapType>(vectorField,a_order,Step)
{}

//###########################################################//


template<typename MapType>
Taylor<MapType>::Taylor(const char* vectorField, int a_order, const ScalarType& Step)
   : BasicTaylor<MapType>(vectorField,a_order,Step)
{}


//###########################################################//

template<typename MapType>
typename Taylor<MapType>::MatrixType Taylor<MapType>::JacPhi(const VectorType &iv) const
{
  // first we compute ceofficients for x
   int r;
   m_coeff[0] = iv;
   for(r=0;r<m_order;++r)
      m_coeff[r+1] = m_vField(m_coeff[r],r)/(r+1);

  // now we compute the coefficients for the Jacobian matrix
  // the coefficients will be stored in F

   m_vField.variational(m_coeff,m_F,m_order);
   // this function compute coefficients up to index 'order'
   computeMatrixCoeff(m_order-1,MatrixType::Identity(m_dimension));

  // the sumation of the Taylor series
   MatrixType result = m_matrixCoeff[m_order];
   for(r=m_order-1;r>=0;--r)
      result = result*m_step + m_matrixCoeff[r];

   return result;
}


//###########################################################//


template<typename MapType>
typename Taylor<MapType>::VectorType Taylor<MapType>::Remainder(const VectorType &iv) const
{
   // we compute an enclosure for \varphi([0,timestep],iv)
   int ok;
   m_coeff[0] = enclosure(iv,&ok);

// the computation of Taylor coefficients up to an order <order+1> on <enc>
   for(int i=0;i<=m_order;++i)
      m_coeff[i+1] = m_vField(m_coeff[i],i)/(i+1);

   return m_coeff[m_order+1]*power(m_step,m_order+1);
}

//###########################################################//


template<typename MapType>
typename Taylor<MapType>::VectorType Taylor<MapType>::enclosure(const VectorType &x) const
// the function finds an enclosure for \varphi([0,step],x)
{
   return capd::dynsys::enclosure(m_vField, x, m_step);
}

//###########################################################//
template<typename MapType>
typename Taylor<MapType>::VectorType Taylor<MapType>::enclosure(const VectorType &x, int *found) const
// the function finds an enclosure for \varphi([0,step],x)
{
  return enclosure(x);
}


//###########################################################//


template<typename MapType>
typename Taylor<MapType>::MatrixType Taylor<MapType>::jacEnclosure(const VectorType &enc, NormType &the_norm) const
// the function finds enclosure for Jacobian matrix (variational part)
// source- "C^1-Lohner algorithm" by P. Zgliczynski
{
   MatrixType der = m_vField[enc], result(m_dimension,m_dimension);

   ScalarType l = the_norm(der).rightBound(); // computation of lagarithmic norm
   ScalarType s = ScalarType(0,1)*m_step;
   ScalarType w = ScalarType(-1,1)*exp(s*l);

   MatrixType W(m_dimension,m_dimension); // W_3 in paper "C^1 - Lohner algorithm"
   W = w;

   result = MatrixType::Identity(m_dimension) + s*der*W;

   int i,j;
   for(i=1;i<=m_dimension;++i)
      for(j=1;j<=m_dimension;++j)
      {
         ScalarType d = result(i,j);
         typename ScalarType::BoundType 
              l = (w.leftBound() > d.leftBound() ? w.leftBound() : d.leftBound()),
              r = (w.rightBound() < d.rightBound() ? w.rightBound() : d.rightBound());
         result(i,j) = ScalarType(l,r);
      }

   return result;
}


//###########################################################//


template<typename MapType>
void Taylor<MapType>::JacRemainder(
      const VectorType &vecEnc,
      const MatrixType &jacEnc,
      VectorType &Rem,
      MatrixType &jacRem
   ) const
// the remainder for varaiational part is computed
   // vecEnc - enclosure for \varphi([0,step],x)
   // jacEnc -  enclosure for V([0,step],x)
{
   m_coeff[0] = vecEnc;
   int r;
   for(r=0;r<=m_order;++r)
      m_coeff[r+1] = m_vField(m_coeff[r],r)/(r+1);

   m_vField.variational(m_coeff,m_F,m_order+1);
   computeMatrixCoeff(m_order,jacEnc);

   ScalarType fac = power(m_step,m_order+1);
   Rem = m_coeff[m_order+1] * fac;
   jacRem = m_matrixCoeff[m_order+1] * fac;
}


//###########################################################//

}} //namespace capd::dynsys

#endif // #define _TAYLOR_HPP

/// @}
