/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file BasicFunction.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2006 */

#ifndef _BasicFunction_HPP
#define _BasicFunction_HPP
#include <sstream>
#include <stdexcept>
#include <cstdlib>

#include "map/BasicFunctionIntf.h"
#include "map/textop.h"

namespace capd{
namespace map{

template<typename Scalar>
void BasicFunction<Scalar>::operator=(const BasicFunction& f)
{
   if(this!=&f)
   {
      deleteTables();
      createFromObject(f);
   }
}

// -----------------------------------------------------------------------------

template<typename Scalar>
void BasicFunction<Scalar>::operator=(const std::string& s)
{
   deleteTables();
   createFromText(s,0);
}

// -----------------------------------------------------------------------------

template<typename Scalar>
void BasicFunction<Scalar>::operator=(const char* s)
{
   deleteTables();
   createFromText(s,0);
}

// -----------------------------------------------------------------------------

template<typename Scalar>
BasicFunction<Scalar>::BasicFunction()
{
   createDefault();
}

// -----------------------------------------------------------------------------

template<typename Scalar>
BasicFunction<Scalar>::BasicFunction(const BasicFunction& f)
{
   createFromObject(f);
}

// -----------------------------------------------------------------------------

template<typename Scalar>
BasicFunction<Scalar>::BasicFunction(const std::string&s, int a_order)
{
   createFromText(s,a_order);
}

// -----------------------------------------------------------------------------

template<typename Scalar>
BasicFunction<Scalar>::BasicFunction(const char* s, int a_order)
{
   createFromText(s,a_order);
}

// -----------------------------------------------------------------------------

template<typename Scalar>
BasicFunction<Scalar>::~BasicFunction()
{
   deleteTables();
}

//--------------------- createDefault ----------------------------------

template<typename Scalar>
void BasicFunction<Scalar>::createDefault()
{
   m_dim = 1;
   m_order = 1;
   m_size = 1;

   m_var = new std::string[1];
   m_val = new ScalarType[1];

   m_var[0] = "";
   m_val[0] = ScalarType(0.);
}

// --------------------- deleteTables ---------------------------------

template<typename Scalar>
void BasicFunction<Scalar>::deleteTables()
{
   delete []m_var;
   delete []m_val;
   m_var = 0; // NULL
   m_val = 0; // NULL
   m_nodes.clear();
}

// -----------------------------------------------------------------------------

template<typename Scalar>
void BasicFunction<Scalar>::createFromText(const std::string &s, int a_order)
{
   m_order = a_order+1;
   m_var = TextOperation::splitVariables("var:",s,m_dim);  /* we read variables and dimension */
   m_size = m_dim*m_order;
   m_val = new ScalarType[m_size];
   for(int j=0;j<m_size;++j)
      m_val[j] = ScalarType(0.);               /* the table of arguments */
}

// -----------------------------------------------------------------------------

template<typename Scalar>
void BasicFunction<Scalar>::createFromObject(const BasicFunction &f)
{
   m_dim = f.m_dim;
   m_size = f.m_size;
   m_order = f.m_order;

   m_var = new std::string[m_dim];
   m_val = new ScalarType[m_size];

   int i;
   for(i=0;i<m_size;++i)
      m_val[i] = f.m_val[i];

   for(i=0;i<m_dim;++i)
      m_var[i] = f.m_var[i];
}

// -----------------------------------------------------------------------------

template<typename Scalar>
void BasicFunction<Scalar>::setOrder(int the_new)
{
   if(m_order==the_new+1) return;
   if (the_new<0)
   {
      std::ostringstream message;
      message <<  "BasicFunction error: invalid argument " << the_new << " in method setOrder";
      throw std::out_of_range(message.str());
   }
   m_size = m_dim*(the_new+1);
   ScalarType *temp = new ScalarType[m_size];
   int i;
   for(i=0;i<m_size;++i)
      temp[i]=ScalarType(0.);

   for(i=0;i<m_dim;++i)
      temp[i*(the_new+1)] = m_val[i*m_order];

   delete []m_val;
   m_val = temp;
   m_order = the_new+1;
}

/* _________________________________________________________________ */

template<typename Scalar>
void BasicFunction<Scalar>::setParameter(const std::string &name, const ScalarType& value)
{
   int cnt=0;
   while( (cnt<m_dim) && (m_var[cnt]!=name) ) ++cnt;

   if(cnt<m_dim)
   {
      m_val[m_order*cnt] = value;
   }
}

/* _________________________________________________________________ */

template<typename Scalar>
std::string BasicFunction<Scalar>::variables() const
{
   std::string result = "var:";
   result += m_var[0];
   for(int i=1;i<m_dim;++i)
      result = result + ","  + m_var[i];
   result += ";";
   return result;
}

/* _____________________________ EQNANAL ______________________________ */

template<typename Scalar>
void BasicFunction<Scalar>::eqnanal(std::string &eq, NodeType **N)
/*
   a recurrent function, which analyze <eq> - a text containing a formula
   and produces a tree representing it.
*/
{
   int hp;
   if (eq.at(0)=='+') eq.erase(0,1); // remove '+' from at the begin of <eq>

   // if <eq> starts with '-', then we insert '0' in front to turn
   // this into an occurence of a subtraction
   TextOperation::removeBrackets(eq);

   typename std::map<std::string,NodeType*>::iterator it = m_nodes.find(eq);
   if(it!=m_nodes.end())
   {
      *N = it->second;
      return;
   }

   if (eq.at(0)=='-') eq.insert(0,"0");

   hp = TextOperation::searchForOperator(eq,'+');
   if(hp)
   {
      std::string left = std::string(eq,0,hp);
      std::string right = std::string(eq,hp+1,std::string::npos);
      *N = new SumNode<ScalarType>(NULL,NULL,m_order);
      eqnanal(left,&((*N)->left));
      eqnanal(right,&((*N)->right));
      (*N)->left->m_links++;
      (*N)->right->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForOperator(eq,'-');
   if(hp)
   {
      std::string left = std::string(eq,0,hp);
      std::string right = std::string(eq,hp+1,std::string::npos);
      *N = new DifNode<ScalarType>(NULL,NULL,m_order);
      eqnanal(left,&((*N)->left));
      eqnanal(right,&((*N)->right));
      (*N)->left->m_links++;
      (*N)->right->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForOperator(eq,'*');
   if(hp)
   {
      std::string left = std::string(eq,0,hp);
      std::string right = std::string(eq,hp+1,std::string::npos);
      *N = new MulNode<ScalarType>(NULL,NULL,m_order);
      eqnanal(left,&((*N)->left));
      eqnanal(right,&((*N)->right));
      (*N)->left->m_links++;
      (*N)->right->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForOperator(eq,'/');
   if(hp)
   {
      std::string left = std::string(eq,0,hp);
      std::string right = std::string(eq,hp+1,std::string::npos);
      *N = new DivNode<ScalarType>(NULL,NULL,m_order);
      eqnanal(left,&((*N)->left));
      eqnanal(right,&((*N)->right));
      (*N)->left->m_links++;
      (*N)->right->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForOperator(eq,'^');
   if(hp)
   {
      std::string left = std::string(eq,0,hp);
      std::string right = std::string(eq,hp+1,std::string::npos);
      *N = new PowNode<ScalarType>(NULL,NULL,m_order);
      eqnanal(left,&((*N)->left));
      eqnanal(right,&((*N)->right));
      (*N)->left->m_links++;
      (*N)->right->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   // there was'nt any operator found, hence we look for functions
   hp = TextOperation::searchForFunction("sqr",eq);
   if(hp)
   {
      std::string left = std::string(eq,hp,std::string::npos);
      *N = new SqrtNode<ScalarType>(NULL,m_order);
      eqnanal(left,&((*N)->left));
      (*N)->left->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForFunction("sin",eq);
   if(hp)
   {
      std::string left = std::string(eq,hp,std::string::npos);
      std::string right = "1";
      *N = new SinNode<ScalarType>(NULL,NULL,m_order);
      eqnanal(left,&((*N)->left));
      eqnanal(right,&((*N)->right));
      (*N)->left->m_links++;
      (*N)->right->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForFunction("cos",eq);
   if(hp)
   {
      std::string left = std::string(eq,hp,std::string::npos);
      std::string right = "1";
      *N = new CosNode<ScalarType>(NULL,NULL,m_order);
      eqnanal(left,&((*N)->left));
      eqnanal(right,&((*N)->right));
      (*N)->left->m_links++;
      (*N)->right->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForFunction("exp",eq);
   if(hp)
   {
      std::string left = std::string(eq,hp,std::string::npos);
      *N = new ExpNode<ScalarType>(NULL,m_order);
      eqnanal(left,&((*N)->left));
      (*N)->left->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   hp = TextOperation::searchForFunction("log",eq);
   if(hp)
   {
      std::string left = std::string(eq,hp,std::string::npos);
      *N = new LogNode<ScalarType>(NULL,m_order);
      eqnanal(left,&((*N)->left));
      (*N)->left->m_links++;
      m_nodes[eq]=*N;
      return;
   }

   // this was'nt functions - now we check for variables
   hp=0;
   while((hp<m_dim)&&(m_var[hp]!=eq)) hp++;

   if(hp<m_dim)
   {
      *N = new VarNode<ScalarType>(&m_val[hp*m_order], m_order, hp);
      return;
   }

   // now we check for  constants
   if ( (std::strtod(eq.c_str(),NULL)) || (eq=="0") )
   {
      *N = new ConsNode<ScalarType>( ScalarType(std::strtod(eq.c_str(),NULL)), m_order);
      //m_nodes[eq]=*N;
      return;
   }

   *N = NULL;

   deleteTables();
   std::string re = "undefined symbol '" + eq + "' in the formula: ";
   re += currentFormula();

   throw std::runtime_error(re);
}

/* ______________________ DEPENDS _____________________________*/

template<typename Scalar>
bool BasicFunction<Scalar>::depends(std::string str, int i) const
/**
   This function checks if 'str' depends on i-th coordinate
*/
{
   int hp;

   TextOperation::removeBrackets(str);
   if((str.at(0)=='-') || str.at(0)=='+')
         str.erase(0,1);
   TextOperation::removeBrackets(str);

   // an analysis of arguments of our function

   hp = TextOperation::searchForOperator(str,'+');
   if(hp) return depends(std::string(str,hp+1,std::string::npos),i) ? true : depends(std::string(str,0,hp),i);

   hp = TextOperation::searchForOperator(str,'-');
   if(hp) return depends(std::string(str,hp+1,std::string::npos),i) ? true : depends(std::string(str,0,hp),i);

   hp = TextOperation::searchForOperator(str,'*');
   if(hp) return depends(std::string(str,hp+1,std::string::npos),i) ? true : depends(std::string(str,0,hp),i);

   hp = TextOperation::searchForOperator(str,'/');
   if(hp) return depends(std::string(str,hp+1,std::string::npos),i) ? true : depends(std::string(str,0,hp),i);

   hp = TextOperation::searchForOperator(str,'^');
   if(hp) return depends(std::string(str,0,hp),i);

   hp = TextOperation::searchForFunction("sqr",str);
   if(hp) return depends(std::string(str,hp,std::string::npos),i);

   hp = TextOperation::searchForFunction("sin",str);
   if(hp) return depends(std::string(str,hp,std::string::npos),i);

   hp = TextOperation::searchForFunction("cos",str);
   if(hp) return depends(std::string(str,hp,std::string::npos),i);

   hp = TextOperation::searchForFunction("exp",str);
   if(hp) return depends(std::string(str,hp,std::string::npos),i);

   hp = TextOperation::searchForFunction("log",str);
   if(hp) return depends(std::string(str,hp,std::string::npos),i);

   hp=0;

   while((hp<m_dim)&&(str!=m_var[hp]))
      hp++;

   if (hp!=i)
         return false;
   else
       return true;
}

// --------------------- isVariable ------------------------------------

template<typename Scalar>
bool BasicFunction<Scalar>::isVariable(std::string &f, int i) const
/**
   this function checks if <f> consist of i-th variable only (plus
   eventually brackets)
*/

{
   TextOperation::removeBrackets(f); // removes external brackets

   int hp=0;
   while((hp<m_dim)&&(f!=m_var[hp])) // looking for a variable
      hp++;

   if (hp!=i)
      return false;
   else
      return true;
}

//--------------------- DIFFANAL ----------------------------------

template<typename Scalar>
std::string BasicFunction<Scalar>::diffanal(std::string &str,int i) const
/**
   a recurrent function creating a partial derivative with respect
   to i-th coordinate
*/
{
   int hp,d1,d2;
   std::string b="";
   TextOperation::removeBrackets(str); // removes external brackets

   if(str.at(0)=='+')
      str.erase(0,1);

   TextOperation::removeBrackets(str); //removes external brackets

   if(str.at(0)=='-')
   {    // (-f)'=0-f'
      str.insert(0,"0");
      return diffanal(str,i);
   }

   TextOperation::removeBrackets(str);

   hp = TextOperation::searchForOperator(str,'+');  // (f+g)'=f'+g'
   if(hp)
   {
      b="";
      std::string left = std::string(str,0,hp);
      std::string right = std::string(str,hp+1,std::string::npos);
      d1 = depends(left,i);   // checks if f depends on i-th variable
      d2 = depends(right,i);   // checks if g depends on i-th variable

      if(d1)
         b = diffanal(TextOperation::removeBrackets(left),i);
      if(d1&&d2)
         b += "+";                           // both arguments depend on i-th variable
      if(d2)
         b += diffanal(TextOperation::removeBrackets(right),i);
      return b;
   }


   hp = TextOperation::searchForOperator(str,'-');    // (f-g)'=f'-g'
   if(hp)
   {                     // we proceed as in case of '+'
      b="";
      std::string left = std::string(str,0,hp);
      std::string right = std::string(str,hp+1,std::string::npos);
      d1 = depends(left,i);   // checks if f depends on i-th variable
      d2 = depends(right,i);   // checks if g depends on i-th variable

      if(d1)
         b += diffanal(TextOperation::removeBrackets(left),i);
      if(d1&&d2)
         b += "-(" + diffanal(TextOperation::removeBrackets(right),i) + ")";
      if((!d1)&&d2)
         b += "0-(" + diffanal(TextOperation::removeBrackets(right),i) + ")"; // then (f-g)'=-(g')
      return b;
   }


   hp = TextOperation::searchForOperator(str,'*');  //(f*g)'=f'*g+f*g'
   if(hp)
   {
      b="";
      std::string left = std::string(str,0,hp);
      std::string right = std::string(str,hp+1,std::string::npos);
      d1 = depends(left,i);   // checks if f depends on i-th variable
      d2 = depends(right,i);   // checks if g depends on i-th variable

      if(d1)
      {               // we need to include  (f')*g
         if (!isVariable(left,i))
            b += "(" + diffanal(TextOperation::removeBrackets(left),i) + ")*";
         b += right;
      }

      if(d2&&d1) b += "+";

      if(d2)
      {               // we need to include f*(g')
         b += "(" + left + ")";
         if(!isVariable(right,i))
            b += "*(" + diffanal(TextOperation::removeBrackets(right),i) + ")";
      }
      return b;
   }


   hp = TextOperation::searchForOperator(str,'/'); // (f/g)'=(f'*g-f*g')/g^2
   if(hp)
   {
      b="";
      std::string left = std::string(str,0,hp);
      std::string right = std::string(str,hp+1,std::string::npos);
      d1 = depends(left,i);   // checks if f depends on i-th variable
      d2 = depends(right,i);   // checks if g depends on i-th variable

      if(d1&&(!d2))      // in this situation (f/g)'=(f')/g
         b += "(" + diffanal(TextOperation::removeBrackets(left),i) + ")/" + right;

      if((!d1)&&d2)
      {     //  (f/g)'= (-(g')*f)/g^2
         b += "(0-";
         if(!isVariable(right,i))
            b += "(" + diffanal(TextOperation::removeBrackets(right),i) + ")*";
         b += left + ")/(" + right + ")^2";
      }

      if(d1&&d2)
      {        // full formula for derivative of f/g
         b += "(" + right;
         if(!isVariable(left,i))
            b += "*(" + diffanal(TextOperation::removeBrackets(left),i) + ")";
         b += "-";
         if(!isVariable(right,i))
            b += "(" + diffanal(TextOperation::removeBrackets(right),i) + ")*(";
         b += left + "))/(" + right + ")^2";
      }
      return b;
   }


   hp = TextOperation::searchForOperator(str,'^');
   if(hp)
   {
      b="";
      std::string left = std::string(str,0,hp);
      std::string right = std::string(str,hp+1,std::string::npos);
      d1 = depends(left,i);   // checks if f depends on i-th variable

      if (!d1)
         return "";
      double j = std::strtod((b=TextOperation::removeBrackets(right)).c_str(),NULL); // check the exponent
      if(j==0) return "0";                              // (f^0)'=0
      if(j==1) return diffanal(TextOperation::removeBrackets(left),i);   // (f^1)'=f'
      if(j==2)
      {                                       // (f^2)'=f*2*(f')
         b="";
         b += left + "*2";
         if(!isVariable(left,i))
            b += "*(" + diffanal(TextOperation::removeBrackets(left),i) + ")";
         return b;
      }

// in remaining cases, also for negative exponents
// (f^n)'=(f')*(n)*f^(n-1)

      b="";
      if(!isVariable(left,i))
         b += "(" + diffanal(TextOperation::removeBrackets(left),i) + ")*";
      b += "(" + right + ")*(" + left + ")^(";
      std::ostringstream bufor;
      bufor << (j-1);
      b += bufor.str() + ")";
      return b;
   }



   hp = TextOperation::searchForFunction("sqr",str); // sqr(f(x))'=0.5*(f'(x))/sqr(f(x))
   if(hp)
   {
      b="";
      std::string right = std::string(str,hp,std::string::npos);
      d1 = depends(right,i);   // checks if f depends on i-th variable

      if(!d1)
         return "";
      b = "";
      b += "0.5*(" + diffanal(TextOperation::removeBrackets(right),i) + ")/" + str; // here str=sqr(...), hence parethesses are not required
      return b;
   }


   hp = TextOperation::searchForFunction("sin",str);
       // sin(x)'=cos(x)  or  sin(f(x))'=cos(f(x))*(f'(x))
   if(hp)
   {
      std::string right = std::string(str,hp,std::string::npos);
      d1 = depends(right,i);   // checks if f depends on i-th variable
      if(!d1)
          return "";
      b = "";
      b += "cos" + right;
      if(!isVariable(right,i))
         b += "*(" + diffanal(TextOperation::removeBrackets(right),i) + ")" ;
      return b;
   }


   hp = TextOperation::searchForFunction("cos",str);
       // cos(x)'=(-sin(x))  or  cos(f(x))'=(-sin(f(x)))*(f'(x))
   if(hp)
   {
      std::string right = std::string(str,hp,std::string::npos);
      d1 = depends(right,i);   // checks if f depends on i-th variable
      if(!d1)
         return "";
      b = ""; b += "(-sin" + right + ")";
      if(!isVariable(right,i))
         b += "*(" + diffanal(TextOperation::removeBrackets(right),i) + ")";
      return b;
   }


   hp = TextOperation::searchForFunction("exp",str);
   // exp(x)'=exp(x)  or  exp(f(x))'=exp(f(x))*(f'(x))
   if(hp)
   {
      std::string right = std::string(str,hp,std::string::npos);
      d1 = depends(right,i);   // checks if f depends on i-th variable
      if(!d1) return "";
      b = "";
      b += str;
      if (!isVariable(right,i))
         b += "*(" + diffanal(TextOperation::removeBrackets(right),i) + ")";
      return b;
   }


   hp = TextOperation::searchForFunction("log",str); // log(f(x))'=(f'(x))/f(x)
   if(hp)
   {
      std::string right = std::string(str,hp,std::string::npos);
      d1 = depends(right,i);   // checks if f depends on i-th variable
      if(!d1) return "";
      b = ""; b += "(" + diffanal(TextOperation::removeBrackets(right),i) + ")/(" + right + ")";
      return b;
   }

   hp=0;
   while((hp<m_dim)&&(m_var[hp]!=str)) hp++;

   if(hp!=i) return "0";
   else return "1";
}

}} // namespace capd::map

#endif // define _BasicFunction_HPP

/// @}
