/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file function.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Daniel Wilczak 2000/2001-2004
   the main function: 'eqnanal' - creates a binary tree
   Each node contains an information about an operator or function:
   +,-,*,/,^ , sin, cos, log, exp , sqr
   the leafs of the tree contain a pointer either to a variable or a constant

   the function is not very sophistcated in the analysis of the formula text,
   hence it is desirable to use many parentheses to clearly indicate the order
   of all operations
*/

#ifndef _FUNCTION_HPP
#define _FUNCTION_HPP

#include <sstream>
#include <string>
#include <stdexcept>

#include "map/nodeintf.h"
#include "map/funcintf.h"
#include "map/textop.h"

namespace capd{
namespace map{


// -----------------------------------------------------------------------

template<typename VectorType>
const std::string& Function<VectorType>::partialDerivative(int i) const
{
   if(m_diff[i]=="?")
   {
      std::string temp = m_fun;
      m_diff[i] = diffanal(temp,i);
   }
   if (m_diff[i]=="") m_diff[i] = "0";
   return m_diff[i];
}

// -----------------------------------------------------------------------

template<typename VectorType>
void Function<VectorType>::setOrder(int a_new)
{
   BasicFunction<ScalarType>::setOrder(a_new);
   m_tree->setOrder(m_order,m_val);
}

/* ______________________________________________________________ */
/*
template<typename VectorType>
const typename Function<VectorType>::ScalarType* const 
   Function<VectorType>::computeDerivatives(const VectorType &v, int a_upTo) const
{
   for(int j=0;j<v.dimension();++j)
      m_val[j*m_order]=v[j];

   m_tree->reset();
   m_tree->computeDerivatives(a_upTo);
   return m_tree->value;
} 
*/
/* _________________________________________________________________ */

template<typename VectorType>
std::string Function<VectorType>::gradient(int numberOfVariables) const
{
   std::string result = variables();
   result += "fun:";
   if(numberOfVariables>m_dim)
   {
      std::ostringstream message;
      message << "Function error: incorrect number of variables requested in gradient method\n";
      message << "Variables: " << variables() << "\n";
      message << "Formula: " << formula() << "\n";
      message << "requested number of variables: " << numberOfVariables;
      throw std::runtime_error(message.str());
   }
   int i;
   for(i=0;i<numberOfVariables-1;++i)
   {
      std::string temp = m_fun;
      if (m_diff[i]=="?")
         m_diff[i] = diffanal(temp,i);
      if (m_diff[i]=="")
         m_diff[i] = "0";
      result += m_diff[i] + ",";
   }

   if (m_diff[numberOfVariables-1]=="?")
   {
      std::string temp = m_fun;
      m_diff[i] = diffanal(temp,i);
   }
   if (m_diff[i]=="")
      m_diff[i] = "0";
   result += m_diff[numberOfVariables-1] + ";";
   return result;
}
}} // namespace capd::map

#endif // #define _FUNCTION_HPP

/// @}
