/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file node.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _NODE_HPP
#define _NODE_HPP

#include <stdexcept>
#include <sstream>
#include <cmath>

#include "map/nodeintf.h"

namespace capd{
namespace map{

template<typename ScalarType>
Node<ScalarType>::~Node(){}

// ---------------------------------------------------------------

template<typename ScalarType>
void Node<ScalarType>::computeDerivatives(int a_upTo)
{
   if(a_upTo>=m_order)
      throw std::runtime_error("Node::computeDerivatives - required m_order to large");
   for(int i=0;i<a_upTo;++i)
      this->operator()(i);
}

// ---------------------------------------------------------------

template<typename ScalarType>
void Node<ScalarType>::reset()
{
   if(m_maxComputedDerivative==-1) return;
   m_maxComputedDerivative = -1;
   if(left) left->reset();
   if(right) right->reset();
}

// ---------------------------------------------------------------

template<typename ScalarType>
ConsNode<ScalarType>::ConsNode(ScalarType a_number,  int a_order) 
   : Node<ScalarType>(a_order,0,0)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   *b = a_number;
   ++b;
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
}

// ---------------------------------------------------------------

template<typename ScalarType>
void ConsNode<ScalarType>::setOrder(int a_order, ScalarType*)
{
   if(m_order==a_order) return;
   ScalarType c = value[0];
   delete [] value;
   m_order = a_order;

   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   *b = c;
   ++b;
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
}

// ---------------------------------------------------------------

template<typename ScalarType>
VarNode<ScalarType>::VarNode(ScalarType *a_number, int a_order, int a_variable)
   : Node<ScalarType>(a_order,0,0)
{
   value = a_number;
   m_numberOfVariables = a_variable;
}

// ---------------------------------------------------------------

template<typename ScalarType>
void VarNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   m_order = a_order;
   value = &a_val[m_numberOfVariables*m_order];
}

// ---------------------------------------------------------------

template<typename ScalarType>
std::ostream& operator<<(std::ostream& out, const capd::map::Node<ScalarType>& n)
{
   out << "{";
   if(n.getOrder()>0){
     if(n[0]>=ScalarType(0)) out << " ";
     out << n[0];
   }
   for(int i=1;i<n.getOrder();i++)
   {
      out << ",";
      if(n[i]>=ScalarType(0)) out << " ";
      out << n[i];
   }
   out << "}";
   return out;
}

}} //namespace capd::map

#endif

/// @}
