/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file basicPoincare.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#include <stdexcept>

#ifndef _BasicPoincare_HPP
#define _BasicPoincare_HPP

namespace capd{
namespace poincare{

// -------------------------------------------------------------------
// the following procedures changes derivatve of the flow onto
// derivative of the Poincare map

template <typename DS>
typename BasicPoincareMap<DS>::VectorType BasicPoincareMap<DS>::computeDT(
      const VectorType& Px,
      const MatrixType& derivativeOfFlow
   )
{
   int j,k;
   VectorType result(m_dim);
   VectorType gradientOnPx = gradient(Px);
   for(j=0;j<m_dim;++j)
      for(k=0;k<m_dim;++k)
         result[j] += gradientOnPx[k] * derivativeOfFlow[k][j];

   result /= - gradientOnPx * dynamicalSystem.getField()(Px);
   return result;
}

// -------------------------------------------------------------------

template <typename DS>
typename BasicPoincareMap<DS>::MatrixType BasicPoincareMap<DS>::computeDP(
      const VectorType& Px,
      const MatrixType& derivativeOfFlow,
      const VectorType& DT
   )
{
   MatrixType result(m_dim,m_dim);
   VectorType fieldOnPx = dynamicalSystem.getField()(Px);
   int i,j;
   for(i=0;i<m_dim;++i)
      for(j=0;j<m_dim;++j)
         result[i][j] = fieldOnPx[i]*DT[j] + derivativeOfFlow[i][j];
   return result;
}

// -------------------------------------------------------------------

template <typename DS>
typename BasicPoincareMap<DS>::MatrixType BasicPoincareMap<DS>::computeD2T(
      const VectorType& Px,
      const MatrixType& derivativeOfFlow,
      const VectorType& DT,
      const MatrixType& DP,
      MatrixType hessianOfFlow[]
   )
{
   int i,j,k,s;
   MatrixType D2T(m_dim,m_dim);
   VectorType fieldOnPx = dynamicalSystem.getField()(Px);
   MatrixType derOfVectorFieldOnPx = dynamicalSystem.getField()[Px];
   VectorType gradientOnPx = gradient(Px);

   for(j=0;j<m_dim;++j)
      for(i=0;i<m_dim;++i)
      {
         ScalarType temp1=0,temp2=0;
         for(k=0;k<m_dim;++k)
         {
            D2T[j][i] -= (gradientOnPx[k] * hessianOfFlow[k][j][i]);
            for(s=0;s<m_dim;++s)
            {
               ScalarType temp3 = gradientOnPx[k] * derOfVectorFieldOnPx[k][s];
               temp1 -= temp3 * DP[s][i];
               temp2 -= temp3 * derivativeOfFlow[s][j];
            }
         }
         D2T[j][i] += DT[j] *temp1 + DT[i]*temp2;
      }
   D2T /= (gradientOnPx * fieldOnPx);
   return D2T;
}

// -------------------------------------------------------------------

template <typename DS>
void BasicPoincareMap<DS>::computeD2P(
         const VectorType& Px,
         const MatrixType& derivativeOfFlow,
         const VectorType& DT,
         const MatrixType& DP,
         MatrixType hessianOfFlow[],
         const MatrixType& D2T,
         MatrixType result[]
        )
{
   int i,j,k,s;
   MatrixType derOfVectorFieldOnPx = dynamicalSystem.getField()[Px];
   VectorType fieldOnPx = dynamicalSystem.getField()(Px);

   for(i=0;i<m_dim;++i)
      for(j=0;j<m_dim;++j)
         for(k=0;k<m_dim;++k)
         {
            result[i][j][k] = fieldOnPx[i] * D2T[j][k] + hessianOfFlow[i][j][k];
            for(s=0;s<m_dim;++s)
            {
               result[i][j][k] += DT[k] * derOfVectorFieldOnPx[i][s] * derivativeOfFlow[s][j];
               result[i][j][k] += DT[j] * derOfVectorFieldOnPx[i][s] * DP[s][k];
            }
         }
}


}} // namespace capd::poincare

#endif // #define _BasicPoincare_HPP

/// @}
