/// @addtogroup poincare
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file poincareIntf.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _PoincareMap_H
#define _PoincareMap_H

// this constant is used as a correction for the time step in procedure
// reachSection. If the computed time for comming close to section is to large
// we multiply it by TIME_STEP_CORECTION_FACTOR. The value 0.9 is just from
// numerical simulations as a good working.
// It can be changed to a value from the interval (0,1)
#define TIME_STEP_CORECTION_FACTOR 0.9

#include <string>
#include "dynset/c0set.h"
#include "dynset/c1set.h"
#include "dynset/c2set.h"
#include "dynset/cnset.h"
#include "poincare/poincex.h"
#include "poincare/basicPoincareIntf.h"

namespace capd{
namespace poincare{

template<typename DS>
class PoincareMap : public BasicPoincareMap<DS>
{
public:
   typedef typename DS::MapType MapType;
   typedef typename DS::FunctionType FunctionType;
   typedef typename DS::MatrixType MatrixType;
   typedef typename DS::VectorType VectorType;
   typedef typename DS::ScalarType ScalarType;
   typedef capd::dynset::C0set<MatrixType> C0set;
   typedef capd::dynset::C1set<MatrixType> C1set;
   typedef capd::dynset::C2set<MatrixType> C2set;
   typedef capd::dynset::CnSet<MatrixType> CnSet;
   typedef typename CnSet::CnCoeffType CnCoeffType;

   PoincareMap(const MapType& dynamicalSystem,
               const FunctionType& section,
               int order,
               const ScalarType& step,
               double factor=0.1
      );

   PoincareMap(const std::string& dynamicalSystem,
               const FunctionType& section,
               int order,
               const ScalarType& step,
               double factor=0.1
      );

   PoincareMap(const char* dynamicalSystem,
               const FunctionType& section,
               int order,
               const ScalarType& step,
               double factor=0.1
      );

// the following procedure is used for C^0 sets
   template<typename T>
   VectorType operator()(T& theSet);

// the following procedure is used for C^1 sets
   template<typename T>
   VectorType operator()(T& theSet, MatrixType& der);

// the following procedure is used for C^2 sets
   template<typename T>
   VectorType operator()(T& theSet, MatrixType& der, MatrixType hessian[]);

// the following procedure is used for C^n sets
   template<typename T>
   VectorType operator()(T& theSet, CnCoeffType& result);

// computations of derivatives of Poincare map from the serivatives of flow
   using BasicPoincareMap<DS>::computeDT;
   using BasicPoincareMap<DS>::computeDP;
   using BasicPoincareMap<DS>::computeD2T;
   using BasicPoincareMap<DS>::computeD2P;

   using BasicPoincareMap<DS>::getField;
   using BasicPoincareMap<DS>::getSection;
   using BasicPoincareMap<DS>::getOrder;
   using BasicPoincareMap<DS>::getStep;

   using BasicPoincareMap<DS>::setField;
   using BasicPoincareMap<DS>::setSection;
   using BasicPoincareMap<DS>::setOrder;
   using BasicPoincareMap<DS>::setStep;

   using BasicPoincareMap<DS>::setFactor;
   using BasicPoincareMap<DS>::setParameter;

protected:
// this function is used to compute the Poincare map
// for all types of C^0, C^1, C^2, C^n sets
   template<typename T>
   void reachSection(T& theSet);

// this function is used to compute the Poincare map
// for all types of C^0, C^1, C^2 and C^n sets
   template<typename T>
   VectorType crossSection(T& theSet);

// these function are used in crossSection only
// the sets of different types should be computed in a different way
// the C^0 sets have not defined the MatrixType operators

   void saveMatrices(C0set&){}
   void saveMatrices(C1set&);
   void saveMatrices(C2set&);
   void saveMatrices(CnSet&);

   void computeJacEnclosure(C0set &S, VectorType &V);
   void computeJacEnclosure(C1set &S, VectorType &V);
   void computeJacEnclosure(C2set &S, VectorType &V);
   void computeJacEnclosure(CnSet &S, VectorType &V);

// the fields of the class

   using BasicPoincareMap<DS>::dynamicalSystem;
   using BasicPoincareMap<DS>::section;
   using BasicPoincareMap<DS>::gradient;
   using BasicPoincareMap<DS>::sectionFactor;
   using BasicPoincareMap<DS>::m_dim;

   MatrixType* derivativeOfFlow;
   MatrixType* hessianOfFlow;
   CnCoeffType* partialDerivatives;
}; // end of template PoincareMap

// -----------------------  inline definitions -------------------------------

template <typename DS>
PoincareMap<DS>::PoincareMap(
      const MapType& dynamicalSystem,
      const FunctionType& section,
      int order,
      const ScalarType& step,
      double factor
   ) : BasicPoincareMap<DS>(dynamicalSystem,section,order,step,factor)
{}

template <typename DS>
PoincareMap<DS>::PoincareMap(
      const std::string& dynamicalSystem,
      const FunctionType& section,
      int order,
      const ScalarType& step,
      double factor
   ) : BasicPoincareMap<DS>(dynamicalSystem,section,order,step,factor)
{}

template <typename DS>
PoincareMap<DS>::PoincareMap(
      const char* dynamicalSystem,
      const FunctionType& section,
      int order,
      const ScalarType& step,
      double factor
   ) : BasicPoincareMap<DS>(dynamicalSystem,section,order,step,factor)
{}

}} // namespace capd::poincare

#endif // define _PoincareMapTemplate

/// @}
