/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file iobject.hpp
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _I_OBJECT_HPP_
#define _I_OBJECT_HPP_

template<typename IntervalObject>
void split(IntervalObject& v, IntervalObject& rv)
{
   v.assertEqualDimension(rv);
   typedef typename IntervalObject::ContainerType ContainerType;
   typename ContainerType::iterator b = v.ContainerType::begin(),e=v.ContainerType::end(), i=rv.ContainerType::begin();
   while(b!=e)
   {
      b->split(*i);
      ++b;
      ++i;
   }
}

template<typename IntervalObject>
typename IntervalObject::ScalarType size(const IntervalObject& v)
{
   typedef typename IntervalObject::ScalarType ScalarType;
   typedef typename IntervalObject::ContainerType ContainerType;
   ScalarType result(0.);
   typename ContainerType::const_iterator b=v.ContainerType::begin(), e=v.ContainerType::end();
   while(b!=e)
   {
      result = capd::max(result,ScalarType(b->rightBound()) - ScalarType(b->leftBound()));
      ++b;
   }
   return ScalarType(result.rightBound());
}

/******************************************************************************/
// checks if IntervalObject v1 is contained in IntervalObject v2
/******************************************************************************/

template<typename IntervalObject>
bool subset(const IntervalObject& v1, const IntervalObject& v2)
{
   v1.assertEqualDimension(v2);
   typedef typename IntervalObject::ContainerType ContainerType;
   typename ContainerType::const_iterator b = v1.ContainerType::begin(), e=v1.ContainerType::end(), i=v2.ContainerType::begin();
   while(b!=e)
   {
      if(!(b->subset(*i))) return false;
      ++b;
      ++i;
   }
   return true;
}

/******************************************************************************/
// checks if IntervalObject v1 is contained in interior of IntervalObject v2
/******************************************************************************/

template<typename IntervalObject>
bool subsetInterior(const IntervalObject& v1, const IntervalObject& v2)
{
   v1.assertEqualDimension(v2);
   typedef typename IntervalObject::ContainerType ContainerType;
   typename ContainerType::const_iterator b = v1.ContainerType::begin(), e=v1.ContainerType::end(), i=v2.ContainerType::begin();
   while(b!=e)
   {
      if(!(b->subsetInterior(*i))) return false;
      ++b;
      ++i;
   }
   return true;
}

/// //////////////////////////////////////////////////////////////////////////////
// intersection
///
/// intersection of two interval objects (vectors, matrices)
/// @returns 
///   true and intersection in result if intersection is not empty
///   false if intersection is empty (value of result is meaningless) 
///
/// //////////////////////////////////////////////////////////////////////////////

template<typename IntervalObject>
bool intersection(const IntervalObject &v1, const IntervalObject &v2, IntervalObject &result)
{
   v1.assertEqualDimension(v2);
   v1.assertEqualDimension(result);
   typedef typename IntervalObject::ContainerType ContainerType;
   typename ContainerType::const_iterator b1 = v1.ContainerType::begin(), b2=v2.ContainerType::begin();
   typename ContainerType::iterator b = result.ContainerType::begin(), e=result.ContainerType::end();

   while(b!=e)
   {
      if( !intersection(*b1,*b2,*b))
      {
         return false;
      }
      ++b;
      ++b1;
      ++b2;
   }
   return true;
}

template<typename IntervalObject>
void intervalHull(const IntervalObject &v1, const IntervalObject &v2, IntervalObject &result)
{
   v1.assertEqualDimension(v2);
   v1.assertEqualDimension(result);
   typedef typename IntervalObject::ContainerType ContainerType;
   typename ContainerType::const_iterator b1 = v1.ContainerType::begin(), b2=v2.ContainerType::begin();
   typename ContainerType::iterator b = result.ContainerType::begin(), e=result.ContainerType::end();

   while(b!=e)
   {
      *b = intervalHull(*b1,*b2);
      ++b;
      ++b1;
      ++b2;
   }
}

template<typename IntervalObject>
void diameter(const IntervalObject &v, IntervalObject &result)
{
   typedef typename IntervalObject::ScalarType ScalarType;
   typedef typename IntervalObject::ContainerType ContainerType;
   typename ContainerType::iterator i = result.ContainerType::begin();
   typename ContainerType::const_iterator b = v.ContainerType::begin(), e=v.ContainerType::end();

   while(b!=e)
   {
      *i = (ScalarType(b->rightBound()) - ScalarType(b->leftBound())).rightBound();
      ++i;
      ++b;
   }
}

template<typename IntervalObject>
void mid(const IntervalObject &v, IntervalObject &result)
{
   typedef typename IntervalObject::ScalarType ScalarType;
   typedef typename IntervalObject::ContainerType ContainerType;
   typename ContainerType::iterator i = result.ContainerType::begin();
   typename ContainerType::const_iterator b = v.ContainerType::begin(), e=v.ContainerType::end();

   while(b!=e)
   {
      *i = b->mid();
      ++i;
      ++b;
   }
}

#endif

/// @}
