/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file matrix.hpp
///
/// @author Marian Mrozek, Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#if !defined(_matrix_hpp_)
#define _matrix_hpp_

#include <vector>
#include <stack>

#include "capd/minmax.h"
#include "vectalg/matrixintf.h"

#include "vectalg/imatrix.hpp"
#include "vectalg/rowVector.hpp"
#include "vectalg/columnVector.hpp"

#include "matrixAlgorithms/floatMatrixAlgorithms.hpp"
#include "matrixAlgorithms/intMatrixAlgorithms.hpp"



//-------------------------- constructors ------------------------//

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>::Matrix(const MatrixSlice<Matrix> &A_m)
   : ContainerType(A_m.numberOfRows(),A_m.numberOfColumns(),true)
{
   if(numberOfRows() && numberOfColumns())
   {
     for (int i=1; i<=numberOfRows(); ++i)
        for (int j=1; j<=numberOfColumns(); ++j)
           (*this)(i,j) = *A_m.at(i,j);
   }
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>::Matrix(Matrix<double,rows,cols>& a_m)
   : ContainerType(a_m.numberOfRows(),a_m.numberOfColumns(),true)
{
   if(numberOfRows() && numberOfColumns())
   {
      typename Matrix<double,rows,cols>::ContainerIterator i=a_m.cbegin();
      ContainerIterator b=cbegin(), e=cend();
      while(b!=e)
      {
         *b = static_cast<Scalar>(*i);
         ++i;
         ++b;
      }
   }
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>::Matrix(int A_rows,int A_cols,const Scalar dane[]) : ContainerType(A_rows,A_cols)
{
   if(numberOfRows() && numberOfColumns())
   {
      int k=0;
      ContainerIterator b=cbegin(), e=cend();
      while(b!=e)
      {
         *b = dane[k];
         ++k;
         ++b;
      }
   }
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>::Matrix(const Scalar dane[]) : ContainerType()
{
   if(numberOfRows() && numberOfColumns())
   {
      int k=0;
      ContainerIterator b=cbegin(), e=cend();
      while(b!=e)
      {
         *b = dane[k];
         ++k;
         ++b;
      }
   }
}

template<typename Scalar, int rows, int cols>
void Matrix<Scalar,rows,cols>::clear()
{
   ContainerIterator b=cbegin(), e=cend();
   while(b!=e)
   {
      *b = Scalar(0);
      ++b;
   }
}

//---------------------------- unary operations -----------------------//

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> abs(const Matrix<Scalar,rows,cols>& v)
{
   Matrix<Scalar,rows,cols> result(v.numberOfRows(),v.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=v.cbegin();

   while(b!=e)
   {
      *b = capd::abs(*i);
      ++b;
      ++i;
   }
   return result;
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator-(const Matrix<Scalar,rows,cols>& v)
{
   Matrix<Scalar,rows,cols> result(v.numberOfRows(),v.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=v.cbegin();

   while(b!=e)
   {
      *b = -(*i);
      ++b;
      ++i;
   }
   return result;
}

//----------------------- arithmetic operations ------------------------//

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator+(const Matrix<Scalar,rows,cols> &a1, const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns())
      throw std::range_error("operator+: incompatible matrix dimensions");

   Matrix<Scalar,rows,cols> result(a1.numberOfRows(),a1.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i1=a1.cbegin(), i2=a2.cbegin();

   while(b!=e)
   {
      *b = (*i1) + (*i2);
      ++b;
      ++i1;
      ++i2;
   }
   return result;
}


template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator-(const Matrix<Scalar,rows,cols> &a1, const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns())
      throw std::range_error("operator-: incompatible matrix dimensions");

   Matrix<Scalar,rows,cols> result(a1.numberOfRows(),a1.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i1=a1.cbegin(), i2=a2.cbegin();

   while(b!=e)
   {
      *b = (*i1) - (*i2);
      ++b;
      ++i1;
      ++i2;
   }
   return result;
}


template<typename Scalar, int rows, int cols1, int cols2>
Matrix<Scalar,rows,cols2> operator*(const Matrix<Scalar,rows,cols1>& a1, const Matrix<Scalar,cols1,cols2>& a2)
{
   Matrix<Scalar,rows,cols2> result(a1.numberOfRows(),a2.numberOfColumns(),true);
   if(a1.numberOfColumns()!=a2.numberOfRows())
      throw std::range_error("operator matrix*matrix: incompatible matrix dimensions");

   for(int i=0;i<result.numberOfColumns();++i)
   {
      typename Matrix<Scalar,rows,cols2>::ContainerIterator b=result.cbegin()+i, e=result.cend()+i;
      typename Matrix<Scalar,rows,cols1>::const_ContainerIterator j=a1.cbegin();
      while(b!=e)
      {
         typename Matrix<Scalar,cols1,cols2>::const_ContainerIterator b1=a2.cbegin()+i, e1=a2.cend()+i;
         Scalar x(0);
         while(b1!=e1)
         {
            x += (*j) * (*b1);
            ++j;
            b1+=a2.rowStride();
         }
         *b=x;
         b+=result.rowStride();
      }
   }

   return result;
}


template<typename Scalar, int rows, int cols>
Vector<Scalar,rows> operator*(const Matrix<Scalar,rows,cols>& a, const Vector<Scalar,cols>& v)
{
   Vector<Scalar,rows> result(a.numberOfRows(),true);
   if(a.numberOfColumns()!=v.dimension())
      throw std::range_error("operator matrix*vector: incompatible dimensions");

   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=a.cbegin();
   typename Vector<Scalar,rows>::iterator b=result.begin(), e=result.end();
   while(b!=e)
   {
      typename Vector<Scalar,cols>::const_iterator bv= v.begin(), be=v.end();
      Scalar x(0);
      while(bv!=be)
      {
         x += (*bv) * (*i);
         ++bv;
         ++i;
      }
      *b=x;
      ++b;
   }
   return result;
}


template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator*(const Matrix<Scalar,rows,cols> &a,const Scalar &s)
{
   Matrix<Scalar,rows,cols> result(a.numberOfRows(),a.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=a.cbegin();

   while(b!=e)
   {
      *b = (*i) * s;
      ++b;
      ++i;
   }
   return result;
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator*(const Scalar &s, const Matrix<Scalar,rows,cols> &a)
{
   Matrix<Scalar,rows,cols> result(a.numberOfRows(),a.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=a.cbegin();

   while(b!=e)
   {
      *b = s * (*i);
      ++b;
      ++i;
   }
   return result;
}


template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator/(const Matrix<Scalar,rows,cols> &a,const Scalar &s)
{
   Matrix<Scalar,rows,cols> result(a.numberOfRows(),a.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=a.cbegin();

   while(b!=e)
   {
      *b = (*i) / s;
      ++b;
      ++i;
   }
   return result;
}


template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator+(const Matrix<Scalar,rows,cols> &a,const Scalar &s)
{
   Matrix<Scalar,rows,cols> result(a.numberOfRows(),a.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=a.cbegin();

   while(b!=e)
   {
      *b = (*i) + s;
      ++b;
      ++i;
   }
   return result;
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> operator-(const Matrix<Scalar,rows,cols> &a,const Scalar &s)
{
   Matrix<Scalar,rows,cols> result(a.numberOfRows(),a.numberOfColumns(),true);
   typename Matrix<Scalar,rows,cols>::ContainerIterator b=result.cbegin(), e=result.cend();
   typename Matrix<Scalar,rows,cols>::const_ContainerIterator i=a.cbegin();

   while(b!=e)
   {
      *b = (*i) - s;
      ++b;
      ++i;
   }
   return result;
}

//------------------------- assignments - matrices --------------------//

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator+=(const Matrix& a)
{
   if(numberOfRows()!=a.numberOfRows() || numberOfColumns()!=a.numberOfColumns() )
      throw std::range_error("operator+=: incompatible matrix dimensions");

   ContainerIterator b=cbegin(),e=cend();
   const_ContainerIterator i=a.cbegin();
   while(b!=e)
   {
      *b += *i;
      ++b;
      ++i;
   }
   return *this;
}


template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator-=(const Matrix& a)
{
   if(numberOfRows()!=a.numberOfRows() || numberOfColumns()!=a.numberOfColumns() )
      throw std::range_error("operator-=: incompatible matrix dimensions");

   ContainerIterator b=cbegin(),e=cend();
   const_ContainerIterator i=a.cbegin();
   while(b!=e)
   {
      *b -= *i;
      ++b;
      ++i;
   }
   return *this;
}

//------------------------ assignments - Scalars ------------------------//

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator=(const Scalar &s)
{
   ContainerIterator b=cbegin(),e=cend();
   while(b!=e)
   {
      *b = s;
      ++b;
   }
   return *this;
}


template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator+=(const Scalar &s)
{
   ContainerIterator b=cbegin(),e=cend();
   while(b!=e)
   {
      *b += s;
      ++b;
   }
   return *this;
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator-=(const Scalar &s)
{
   ContainerIterator b=cbegin(),e=cend();
   while(b!=e)
   {
      *b -= s;
      ++b;
   }
   return *this;
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator*=(const Scalar &s)
{
   ContainerIterator b=cbegin(),e=cend();
   while(b!=e)
   {
      *b *= s;
      ++b;
   }
   return *this;
}

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols>& Matrix<Scalar,rows,cols>::operator/=(const Scalar &s)
{
   ContainerIterator b=cbegin(),e=cend();
   while(b!=e)
   {
      *b /= s;
      ++b;
   }
   return *this;
}

//----------- coord-wise inequalities - true if true for each coord -------//

template<typename Scalar,int rows, int cols>
bool operator<(const Matrix<Scalar,rows,cols> &a1,const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns() )
      throw std::range_error("operator<: incompatible matrix dimensions");

   typename Matrix<Scalar,rows,cols>::const_ContainerIterator b1=a1.cbegin(), b2=a2.cbegin(), e1=a1.cend();
   while(b1!=e1)
   {
      if(!(*b1 < *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int rows, int cols>
bool operator>(const Matrix<Scalar,rows,cols> &a1,const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns() )
      throw std::range_error("operator>: incompatible matrix dimensions");

   typename Matrix<Scalar,rows,cols>::const_ContainerIterator b1=a1.cbegin(), b2=a2.cbegin(), e1=a1.cend();
   while(b1!=e1)
   {
      if(!(*b1 > *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int rows, int cols>
bool operator<=(const Matrix<Scalar,rows,cols> &a1,const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns() )
      throw std::range_error("operator<=: incompatible matrix dimensions");

   typename Matrix<Scalar,rows,cols>::const_ContainerIterator b1=a1.cbegin(), b2=a2.cbegin(), e1=a1.cend();
   while(b1!=e1)
   {
      if(!(*b1 <= *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int rows, int cols>
bool operator>=(const Matrix<Scalar,rows,cols> &a1,const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns() )
      throw std::range_error("operator>=: incompatible matrix dimensions");

   typename Matrix<Scalar,rows,cols>::const_ContainerIterator b1=a1.cbegin(), b2=a2.cbegin(), e1=a1.cend();
   while(b1!=e1)
   {
      if(!(*b1 >= *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


template<typename Scalar,int rows, int cols>
bool operator==(const Matrix<Scalar,rows,cols> &a1,const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns() )
      throw std::range_error("operator==: incompatible matrix dimensions");

   typename Matrix<Scalar,rows,cols>::const_ContainerIterator b1=a1.cbegin(), b2=a2.cbegin(), e1=a1.cend();
   while(b1!=e1)
   {
      if(!(*b1 == *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}

template<typename Scalar,int rows, int cols>
bool operator!=(const Matrix<Scalar,rows,cols> &a1,const Matrix<Scalar,rows,cols> &a2)
{
   if(a1.numberOfRows()!=a2.numberOfRows() || a1.numberOfColumns()!=a2.numberOfColumns() )
      throw std::range_error("operator!=: incompatible matrix dimensions");

   typename Matrix<Scalar,rows,cols>::const_ContainerIterator b1=a1.cbegin(), b2=a2.cbegin(), e1=a1.cend();
   while(b1!=e1)
   {
      if(!(*b1 != *b2))
         return false;
      ++b1;
      ++b2;
   }
   return true;
}


//------------- operations on matrices -------------------------//

template<typename Scalar, int rows, int cols>
Matrix<Scalar,rows,cols> Matrix<Scalar,rows,cols>::Identity(int A_dimension)
{
   // the matrix is filled by zeros
   Matrix<Scalar,rows,cols> temp(A_dimension,A_dimension);
   for(int i=1;i<=A_dimension;++i)
      temp(i,i) = Scalar(1);
   return temp;
}


template<typename Scalar, int rows, int cols>
void Matrix<Scalar,rows,cols>::setToIdentity()
{
   if(numberOfRows()!=numberOfColumns())
      throw std::range_error("Matrix<Scalar,rows,cols>::setToIdentity: rows!=cols");
   clear();
   for(int i=1;i<=numberOfRows();++i)
      (*this)(i,i) = Scalar(1);
}


template<typename Scalar, int rows, int cols>
void Matrix<Scalar,rows,cols>::Transpose()
{
   if(numberOfRows()!=numberOfColumns())
      throw std::runtime_error("Cannot call x.Transpose() for nonsquare matrix");

   for(int i=1;i<=numberOfRows();++i)
   for(int j=i+1;j<=numberOfRows();++j)
   {
      Scalar a = (*this)(i,j);
      (*this)(i,j) = (*this)(j,i);
      (*this)(j,i) = a;
   }
}


template<typename Scalar, int rows, int cols>
Matrix<Scalar,cols,rows> Transpose(const Matrix<Scalar,rows,cols> &a)
{
   Matrix<Scalar,cols,rows> temp(a.numberOfColumns(),a.numberOfRows(),true);
   for(int i=1;i<=a.numberOfColumns();++i)
      for(int j=1;j<=a.numberOfRows();++j)
      {
         temp(i,j)= a(j,i);
      }
   return temp;
}

// ----------------- input - output ------------------------------ //

template<typename Scalar, int rows, int cols>
std::ostream &operator<<(std::ostream &out,const Matrix<Scalar,rows,cols> &a)
{
   out << "{";
   if(a.numberOfColumns()>0){
     if(a.numberOfRows()>1) out << std::endl;
     if(a.numberOfRows()>0) out << a[0];
     for(int i=1;i<a.numberOfRows();++i)
     {
        out << "," << std::endl << a[i];
     }
     if(a.numberOfRows()>1) out << std::endl;
   }
   out << "}";
   return out;
}


template<typename Scalar, int rows, int cols>
std::istream &operator>>(std::istream &inp, Matrix<Scalar,rows,cols> &a){
  typedef typename Matrix<Scalar,rows,cols>::RowVectorType VectorType;
  std::stack<VectorType> st;
  VectorType v(cols);
  int ch=0,dimension=0;
  while('{'!=(ch=inp.get()) && ch!=EOF);
  if(ch!=EOF){
    while(' '==(ch=inp.peek()) || ch=='\n') inp.get();
    if(ch!='}'){ // otherwise we have an ampty matrix
      inp >> v;
      st.push(v);
      dimension=v.dimension();
      do{
        do{
          ch=inp.get();
        }while(ch==' ' || ch=='\t' || ch=='\n');
        if(ch==','){
          inp >> v;
          if(dimension!=v.dimension())
            throw std::ios_base::failure("Rows of incompatible dimensions found when reading a matrix");
          st.push(v);
        }
      }while(ch!='}' && ch!=EOF);
    }
  }

  int n = static_cast<int>(st.size());
  a.resize(n,dimension);
  for(int i=n-1;i>=0;--i)
  {
    v=st.top();
    a[i]=v;
    st.pop();
  }

  if(inp.eof())
    throw std::ios_base::failure("EOF encountered when reading a matrix");
  return inp;
}

#endif //_matrix_hpp_

/// @}
