#include <map>

// this file contains rigorous routines used in the proof of the existence 
// of a branch of period two points connecting first and second period doubling bifurcations
// this file in fact implements Algorithm 2 from the paper, which uses Krawczyk Method 
// for proving the existence of period two point branch


// ----------------------------------------------------------------------------
// this is rigorous routine 
// it verifes that for parameter range [a2,a3] there exists unique 
// period two point in the set 
// (fp1+box1, fp3+box3)
// where fp is an approximate period two point
// The proof uses Interval Krawczyk Operator. The details can be found in Algorithm 2

static int counter=0;
static std::map< int,IVector> *branchTab;
std::map< int,IVector> branch;

template<typename PoincareMapType>
void connectingBranch(
      BasicPoincareMap* BPM[], 
      PoincareMapType* PM[], 
      double a2, double a3, int grid,
      const typename PoincareMapType::VectorType& box1,
      const typename PoincareMapType::VectorType& box2,
      const typename PoincareMapType::VectorType& box3,
      const typename PoincareMapType::VectorType& box4,
      const DVector& startPoint
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

// divide the interval [a2,a3] onto grid parts
   LInterval skipA = (LInterval(a3) - LInterval(a2))/grid;
   LInterval boxA = LInterval(0,1)*skipA;
   LMatrix Id = LMatrix::Identity(2);

   int i=0;
   #pragma omp  parallel default(none) shared(startPoint,box1,box2,box3,box4,BPM, PM, skipA, std::cout, boxA,grid,a2,a3,counter,branchTab) private(i)
   {
      #pragma omp for nowait
      for(i=0;i<grid;i++)
      {
         LInterval a = LInterval(a2) + i*skipA + boxA;
         PM[omp_get_thread_num()]->setParameter("a",a);
         BPM[omp_get_thread_num()]->setParameter("a",a2+i*(a3-a2)/grid);
         // find approximate period two point using nonrigirous Newton method
         // this is a good candidate for the center of the set on which we verify existence 
         // and uniqueness of period two point
         DVector fp1 = findPeriodicOrbit(*BPM[omp_get_thread_num()],startPoint,4,18);
         DVector fp2(fp1);
         computePoincare(*BPM[omp_get_thread_num()],1,fp2);
         DVector fp3(fp2);
         computePoincare(*BPM[omp_get_thread_num()],1,fp3);
         DVector fp4(fp3);
         computePoincare(*BPM[omp_get_thread_num()],1,fp4);

        // create an interval vector from the 'double' vector
         LVector sP1(2), sP2(2), sP3(2), sP4(2);
         sP1[0]=fp1[1];
         sP1[1]=fp1[2];
         sP2[0]=fp2[1];
         sP2[1]=fp2[2];
         sP3[0]=fp3[1];
         sP3[1]=fp3[2];
         sP4[0]=fp4[1];
         sP4[1]=fp4[2];

         // compute IntervalKrawczykOperator
         // the procedure verifyKrawczykOperator throws an exception if inclusion is not satisfied
         LVector N = verifyKrawczykOperator(*PM[omp_get_thread_num()],sP1,sP2,sP3,sP4,box1,box2,box3,box4);
         // so here we are sure that N contains period two point

         // optional, one can uncomment it
         //std::cout << "step " << (i+1) << " of " << grid << std::endl;
         //std::cout << "a: " << a <<  std::endl;
         //std::cout << "N:" << N << std::endl;

         // we store the result in the memory
         branchTab[omp_get_thread_num()][counter+i] = N;
      }
   }
   counter += grid;
}

// ----------------------------------------------------------------------------
// the following procedure calls 74 times the previous one with parameter values listed in Table 5 of the paper

void connectingBranch(BasicPoincareMap* BPM[], PoincareMap* PM[], MainData& md1, MainData& md2)
{
   double a3 = 3.83735812;

   IVector box1(2), box2(2), box3(2),box4(2);
   DVector startPoint(0.,-4.5,0.06);

   branchTab = new std::map< int,IVector>[omp_get_max_threads()];
   int i;
   for(i=0;i<omp_get_max_threads();++i)
   {
      PM[i]->setStep(interval(0.03));
      PM[i]->setOrder(7);
   }

   box1[0] = interval(-1.,1.)*1.1e-5;
   box1[1] = interval(-1.,1.)*3e-8;
   box2[0] = interval(-1.,1.)*1e-5;
   box2[1] = interval(-1.,1.)*3e-5;
   box3[0] = interval(-1.,1.)*1.1e-5;
   box3[1] = interval(-1.,1.)*3e-8;
   box4[0] = interval(-1.,1.)*1e-5;
   box4[1] = interval(-1.,1.)*3e-5;

   connectingBranch(BPM,PM,2.83290, 2.83291,14000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83291, 2.83292,13500,box1,box2,box3,box4,startPoint);

   box1 *= 3;
   box2 *= 3;
   box3 *= 3;
   box4 *= 3;  
   connectingBranch(BPM,PM,2.83292, 2.83293,12200,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83293, 2.83294,11250,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83294, 2.83295,10400,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83295, 2.83296,9650,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83296, 2.83297,9050,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83297, 2.83298,8500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83298, 2.83299,8000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83299, 2.8330,7550,box1,box2,box3,box4,startPoint);

   connectingBranch(BPM,PM,2.833, 2.83301,7200,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83301, 2.83302,6800,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83302, 2.83303,6500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83303, 2.83304,6200,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83304, 2.83305,6000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83305, 2.83306,5700,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83306, 2.83307,5500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83307, 2.83308,5300,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83308, 2.83309,5100,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83309, 2.8331,4900,box1,box2,box3,box4,startPoint);

   connectingBranch(BPM,PM,2.8331, 2.83311,4800,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83311, 2.83312,4600,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83312, 2.83313,4450,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83313, 2.83314,4300,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83314, 2.83315,4150,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83315, 2.83316,4050,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83316, 2.83317,3950,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83317, 2.83318,3850,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83318, 2.83319,3750,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.83319, 2.8332,3650,box1,box2,box3,box4,startPoint);


   connectingBranch(BPM,PM,2.8332, 2.8333,36000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8333, 2.8334,29000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8334, 2.8335,24000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8335, 2.8336,20000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8336, 2.8337,18000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8337, 2.8338,16000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8338, 2.8339,14000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8339, 2.834,13000,box1,box2,box3,box4,startPoint);

   connectingBranch(BPM,PM,2.834, 2.8345,59000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8345, 2.835,42000,box1,box2,box3,box4,startPoint);

   box1 *= 5;
   box2 *= 5;
   box3 *= 5;
   box4 *= 5;

   connectingBranch(BPM,PM,2.835, 2.8355,17500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8355, 2.836,11000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.836, 2.8365,8000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8365, 2.837,6200,box1,box2,box3,box4,startPoint);

   box1 *= 2;
   box2 *= 2;
   box3 *= 2;
   box4 *= 2;

   connectingBranch(BPM,PM,2.837, 2.8372,2800,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8372, 2.8375,3600,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8375, 2.838,5000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.838, 2.8385,3600,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8385, 2.839,2900,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.839, 2.8395,2400,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.8395, 2.84,2100,box1,box2,box3,box4,startPoint);

   connectingBranch(BPM,PM,2.84, 2.841,3700,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.841, 2.842,3000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.842, 2.843,2500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.843, 2.844,2200,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.844, 2.845,2000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.845, 2.846,1700,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.846, 2.848,3100,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.848, 2.85,2600,box1,box2,box3,box4,startPoint);

   connectingBranch(BPM,PM,2.85, 2.86,11000,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.86, 2.87,6500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.87, 2.88,4700,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.88, 2.89,3700,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.89, 2.9,3100,box1,box2,box3,box4,startPoint);

   box1 *= 5;
   box2 *= 5;
   box3 *= 5;
   box4 *= 5;

   connectingBranch(BPM,PM,2.9, 2.95,4400,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,2.95, 3.0,2500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.0, 3.1,3500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.1, 3.2,2500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.2, 3.3,2100,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.3, 3.4,1800,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.4, 3.5,1600,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.5, 3.6,1400,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.6, 3.7,1500,box1,box2,box3,box4,startPoint);
   connectingBranch(BPM,PM,3.7, a3,2400,box1,box2,box3,box4,startPoint);

   for(i=0;i<omp_get_max_threads();++i)
   {
      branch.insert(branchTab[i].begin(),branchTab[i].end());
      branchTab[i].clear();
   }

   // at the end we must verify if the branch is connected set
   for(unsigned int i=0;i<branch.size()-1;++i)
   {
      // the procedure intersection throws an exception if this intersection is empty
      intersection(branch[i],branch[i+1]);
   }

// finally, we verify if the branches of period two points glue to continuous curve
   IVector box(2);
   box[0]=interval(-2.87757e-2,2.79e-2);
   box[1]=interval(-8e-4,8e-4);

// at swiching point a = a_s
   IVector r = gauss(md1.ICoordSystem,branch[0]-md1.Ix);
   if(!subsetInterior(r,box))
      throw std::runtime_error("Cannot verify if branches of period two point glue to coninuous curve at a=a_s");

// at a=a3

   IVector B  = branch[branch.size()-1];
   if(!subsetInterior(md2.Ix + md2.ICoordSystem*md2.box,B))
   {
      throw std::runtime_error("Cannot verify if branches of period two point glue to coninuous curve at a=a_3");
   }
}
