// ----------------------------------------------------------------------------

// this procedure computes a rigorous image of a set and its partial derivatives 
// the initial condition is represented as 
// x + C*r
// iter - number of iterations of Poincare map
// result - a structure which stores partial derivatives of a map
template<typename PoincareMapType>
typename PoincareMapType::VectorType computePoincareMap(
      PoincareMapType& PM, 
      const typename PoincareMapType::VectorType& x, 
      const typename PoincareMapType::MatrixType& C, 
      const typename PoincareMapType::VectorType& r, 
      int iter, 
      typename PoincareMapType::CnCoeffType& result
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

// rank - order of partial derivatives
   int rank = result.rank();
   int dim = x.dimension();

// we set 0 at the first coordinate, this is the Poincare section
   LVector copyX(dim+1);
   LVector copyR(dim+1);
   LMatrix copyC(dim+1,dim+1);

   copyR[0] = copyX[0] = LInterval(0.); // Poincare section
   copyC(1,1) = LInterval(1.);

   int i,j;
   for(i=1;i<=dim;++i)
   {
      copyX[i] = x[i-1];
      copyR[i] = r[i-1];
      for(j=1;j<=dim;++j)
         copyC(i+1,j+1) = C(i,j);
   }
   
// computing iter-iteration of a flow
   capd::dynset::CnRect2<LMatrix> theSet(copyX,copyC,copyR,EuclLNorm<LVector,LMatrix>(),rank);
   LCnCoeff image(dim+1,rank);
   for(int i=1;i<=iter;++i)
   {
      PM(theSet,image);
   }
   // recompute partial derivatives of Poincare map from partial derivatives of the flow
   image = PM.computePartialDerivatives(image);

   LVector res(dim);
   for(i=1;i<=dim;++i)
      res[i-1]=image(i);
   
// copying result from 'image' object to 'result' object (result has not first coordinate which defines poincare section)
   for(int s=1;s<=rank;++s)
   {
      capd::vectalg::Multipointer mp = result.first(s);
      do{
         capd::vectalg::Multipointer mp2=mp;
         for(j=0;j<mp2.dimension();++j)
            ++mp2[j];
         for(i=1;i<=dim;++i)
            result(i-1,mp) = image(i,mp2);
      }while(result.next(mp));
   }

   return res;
}

// ----------------------------------------------------------------------------

// this procedure computes a rigorous image of a set
// the initial condition is represented as 
// x + C*r
// iter - number of iterations of Poincare map
template<typename PoincareMapType>
typename PoincareMapType::VectorType computePoincareMap(
      PoincareMapType& PM, 
      const typename PoincareMapType::VectorType& x, 
      const typename PoincareMapType::MatrixType& C, 
      const typename PoincareMapType::VectorType& r, 
      int iter
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

   int dim = x.dimension();

// we set 0 at the first coordinate, this is the Poincare section
   LVector copyX(dim+1);
   LVector copyR(dim+1);
   LMatrix copyC(dim+1,dim+1);

   copyR[0] = copyX[0] = LInterval(0.); // Poincare section
   copyC(1,1) = LInterval(1.);

   int i,j;
   for(i=1;i<=dim;++i)
   {
      copyX[i] = x[i-1];
      copyR[i] = r[i-1];
      for(j=1;j<=dim;++j)
         copyC(i+1,j+1) = C(i,j);
   }
   
   capd::dynset::Rect2Set<LMatrix> theSet(copyX,copyC,copyR);
   LVector res(dim+1);
   for(int i=0;i<iter;++i)
      res = PM(theSet);

   return take2DVector(res);
}

// ----------------------------------------------------------------------------
// this is a rigorous routine
// it proves, that in the set x + coordSystem*box there exists a unique fixed point for P^2
// if an inlusion required in the interval Newton operator is not satisfied, the procedure throws an exception 
template<typename PoincareMapType>
typename PoincareMapType::VectorType provePeriodicOrbit(
      PoincareMapType& PM, 
      const typename PoincareMapType::VectorType& x, 
      const typename PoincareMapType::MatrixType& coordSystem, 
      const typename PoincareMapType::MatrixType& invCoordSystem, 
      typename PoincareMapType::VectorType box, 
      int iter
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

   // computation of P^2(x) at the center
   LVector Fcenter = computePoincareMap(PM,x,coordSystem,LVector(2)/*this is zero*/,iter);
   Fcenter = changeCoordSystem(x,invCoordSystem,Fcenter);

   // computation of DP^2(X)
   LCnCoeff der(2,1); // two variables, C^1 computations
   computePoincareMap(PM,x,coordSystem,box,iter,der);
   changeCoordSystem(x,coordSystem,invCoordSystem,Fcenter,der);

   LMatrix DP = LMatrix(der) - LMatrix::Identity(2);
   // the interval Newtion operator
   LVector Nx = -inverse2x2Matrix(DP)*Fcenter;
   
   if(!subsetInterior(Nx,box))
   {
      std::cout << "Nx: " << Nx << std::endl;
      std::cout << "r: " << box << std::endl << std::endl;
      throw std::runtime_error("Interval Newton Operator: inclusion not satisfied");
   }
   return Nx; // the fixed point is in x+coordSystem*Nx
}


// ----------------------------------------------------------------------------
// this is a rigorous routine. It verifies that for given a and x there exists unique
// y in box[1] such that P^2(a,x,y)=y
// the function returns computed y
template<typename PoincareMapType, typename MainDataType>
typename PoincareMapType::ScalarType findYOfX(
      PoincareMapType& PM, 
      typename PoincareMapType::ScalarType a, 
      typename PoincareMapType::ScalarType x, 
      typename PoincareMapType::ScalarType Iy, 
      MainDataType& md
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

   LVector center(2);
   center[0] = x.mid();
   center[1]= Iy.mid();
   LVector box(2);
   box[0]=x-x.mid();
   box[1] = 0;

   PM.setParameter("a",a);
   LVector image = computePoincareMap(PM,md.Ix+md.ICoordSystem*center,md.ICoordSystem,box,4*md.bifNumber);
   image = changeCoordSystem(md.Ix,md.IInvCS,image);
   LInterval Fy = image[1]; // y-F(y), y=box[1]=0

   box[1] = Iy-Iy.mid();
   LCnCoeff DF(2,1);
   image = computePoincareMap(PM,md.Ix+md.ICoordSystem*center,md.ICoordSystem,box,4*md.bifNumber,DF);
   changeCoordSystem(md.Ix,md.ICoordSystem,md.IInvCS,image,DF);
   
   LInterval Ny = Fy/(1-DF(1,1));
   if(!Ny.subsetInterior(Iy))
   {
      std::cout << "Ny: " << Ny << std::endl;
      std::cout << "Iy: " << Iy << std::endl;
      throw std::runtime_error("findYOfX: inclusion not satisfied");
   }
   if(!Ny.subset(md.box[1]))
   {
      std::cout << "Ny:     " << Ny << std::endl;
      std::cout << "box[1]: " << md.box[1] << std::endl;      
      throw std::runtime_error("findYOfX: Nx is not a subset of box[1]");
   }
   return Ny;
}

// ----------------------------------------------------------------------------
// this is a rigorous routine. It verifies that for given a and x there exists unique
// y in box[1] such that P^2(a,x,y)=y
// the function returns computed y 
// this function uses the algorithm given above and interates twice IntervalNewtonOperator

template<typename PoincareMapType, typename MainDataType>
typename PoincareMapType::ScalarType findYOfX(
      PoincareMapType& PM, 
      typename PoincareMapType::ScalarType a, 
      typename PoincareMapType::ScalarType x, 
      MainDataType& md, bool findBetterEnclosure=true
   )
{
   typename PoincareMapType::ScalarType Ny = findYOfX(PM,a,x,md.box[1],md);
   if(findBetterEnclosure)
   {
      try{
            typename PoincareMapType::ScalarType betterY = findYOfX(PM,a,x,Ny,md);
            Ny = betterY;
      }catch(...){} // if cannot find better enclosure we proceed with previously computed
   }
   return Ny;
}

// ----------------------------------------------------------------------------

// this is a rigorous routine
// it proves, that in the sets x1+box1 x3+box3 there exists a unique period two point for P^2
// i.e. a\in x1+box1 P^2(a)\in x3+box3 and P^2(P^2(a)) = a 
// moreover, the sets x1+box1 and x3+box3 are disjoint
// if an inlusion required in the interval Newton operator is not satisfied, the procedure throws an exception 

template<typename PoincareMapType>
typename PoincareMapType::VectorType verifyKrawczykOperator(
      PoincareMapType& PM, 
      const typename PoincareMapType::VectorType& x1, 
      const typename PoincareMapType::VectorType& x2, 
      const typename PoincareMapType::VectorType& x3, 
      const typename PoincareMapType::VectorType& x4, 
      typename PoincareMapType::VectorType box1, 
      typename PoincareMapType::VectorType box2, 
      typename PoincareMapType::VectorType box3, 
      typename PoincareMapType::VectorType box4
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

   // we need to check if it is period two orbit, i.e. x1+box1 \cap x3+box3 = \emptyset
   LVector s1 = x1+box1;
   LVector s2 = x2+box2;
   LVector s3 = x3+box3;
   LVector s4 = x4+box4;
   if( !(s1[0]>s3[0]) )
      throw std::runtime_error("verifyKrawczykOperator: the sets are not disjoint");
      
   LMatrix Id = LMatrix::Identity(x1.dimension());
   // computation of P^2(x)
   LVector Fcenter1 = x2 - computePoincareMap(PM,x1,Id,LVector(x1.dimension()),1);
   LVector Fcenter2 = x3 - computePoincareMap(PM,x2,Id,LVector(x1.dimension()),1);
   LVector Fcenter3 = x4 - computePoincareMap(PM,x3,Id,LVector(x1.dimension()),1);
   LVector Fcenter4 = x1 - computePoincareMap(PM,x4,Id,LVector(x1.dimension()),1);

   LVector Fcenter(8), box(8);
   box[0]=box1[0]; box[1] = box1[1]; 
   box[2]=box2[0]; box[3] = box2[1];
   box[4]=box3[0]; box[5] = box3[1];
   box[6]=box4[0]; box[7] = box4[1];
   
   Fcenter[0] = Fcenter1[0]; Fcenter[1] = Fcenter1[1]; 
   Fcenter[2] = Fcenter2[0]; Fcenter[3] = Fcenter2[1]; 
   Fcenter[4] = Fcenter3[0]; Fcenter[5] = Fcenter3[1]; 
   Fcenter[6] = Fcenter4[0]; Fcenter[7] = Fcenter4[1]; 

   // computation of DP^2(X)
   LCnCoeff der1(x1.dimension(),1), der2(x1.dimension(),1), der3(x1.dimension(),1), der4(x1.dimension(),1);
   computePoincareMap(PM,x1,Id,box1,1,der1);
   computePoincareMap(PM,x2,Id,box2,1,der2);
   computePoincareMap(PM,x3,Id,box3,1,der3);
   computePoincareMap(PM,x4,Id,box4,1,der4);

   LMatrix DF(8,8);
   DF(1,1) = -der1(0,0); DF(1,2) = -der1(0,1);
   DF(2,1) = -der1(1,0); DF(2,2) = -der1(1,1);

   DF(3,3) = -der2(0,0); DF(3,4) = -der2(0,1);
   DF(4,3) = -der2(1,0); DF(4,4) = -der2(1,1);
   
   DF(5,5) = -der3(0,0); DF(5,6) = -der3(0,1);
   DF(6,5) = -der3(1,0); DF(6,6) = -der3(1,1);

   DF(7,7) = -der4(0,0); DF(7,8) = -der4(0,1);
   DF(8,7) = -der4(1,0); DF(8,8) = -der4(1,1);

   DF(1,3) = DF(2,4) = DF(3,5) = DF(4,6) = DF(5,7) = DF(6,8) = DF(7,1) = DF(8,2) = LInterval(1.);

   LMatrix C = midMatrix(inverseMatrix(midMatrix(DF)));
   // the interval Krawczyk operator
   LVector Nx = -C*Fcenter + (LMatrix::Identity(8)-C*DF)*box;

   // in fact KrawczykOperator = (x1,x2,x3,x4)+Nx and it should be a subset of (x1,x2,x3,x4)+box
   // it is enough to verify Nx\subset box
   if(!subsetInterior(Nx,box))
   {
      std::cout << "Nx: " << Nx << std::endl;
      std::cout << "box : " << box << std::endl << std::endl;
      throw std::runtime_error("Interval Krawczyk Operator: inclusion not satisfied");
   }
   // the period two point is in the set x1+Nx[1]:
   LVector result(2);
   result[0] = x1[0] + Nx[0];
   result[1] = x1[1] + Nx[1];
   
   return result;
}
