// ----------------------------------------------------------------------------
// this is rigorous routine
// it computes derivatives \partial^3 G/ \partial x^3 and \partial^2 G/ \partial x^2

void computeD3G(PoincareMap& PM, interval a, interval x, const MainData& md, interval& D2G, interval& D3G, bool findBetterEnclosure)
{
   interval y = findYOfX(PM,a,x,md,findBetterEnclosure);

   IVector u(2);
   u[0]=x; u[1]=y;
   IVector center = midVector(u);
   IVector box = u-center;

   CnCoeff DP2(2,3);
   IVector image = computePoincareMap(PM,md.Ix + md.ICoordSystem*center,md.ICoordSystem,box,4*md.bifNumber,DP2);
   image = changeCoordSystem(md.Ix,md.ICoordSystem,md.IInvCS,image,DP2);

   interval temp = 1-DP2(1,1);
   interval   dy = DP2(1,0)/temp;
   interval  ddy = ( (DP2(1,1,1)*dy + 2*DP2(1,0,1))*dy + DP2(1,0,0) )/temp;
   interval dddy = ( ((DP2(1,1,1,1)*dy + 3*DP2(1,0,1,1))*dy + 3*(ddy*DP2(1,1,1)+DP2(1,0,0,1)))*dy + 3*ddy*DP2(1,0,1) + DP2(1,0,0,0))/temp;

   D2G = -DP2(0,1,1)*(dy^2) - 2*DP2(0,0,1)*dy - ddy* DP2(0,1) - DP2(0,0,0);

   D3G  = dddy*DP2(0,1);
   D3G -= dy*(3*ddy*DP2(0,1,1) + dy*(dy*DP2(0,1,1,1) + 3*DP2(0,0,1,1)) + 3*DP2(0,0,0,1));
   D3G -= 3*ddy*DP2(0,0,1);
   D3G -= DP2(0,0,0,0);
}


// ----------------------------------------------------------------------------
// this is rigorous routine
// it computes x_fp'(a). We assume that where (x,y) is a set which contains a fixed point curve (x_fp(a),y(a,x_fp(a)))

interval computeDXA(PoincareMap& PM4D, const interval& x, const interval& y, const interval& a, const MainData& md)
{
// here we need C^1 integration in four dimensional space
   IVector u(3);
   u[0]=x; u[1]=y; u[2]=a;
   IVector center = midVector(u);
   IVector box = u-center;

   IMatrix CS = expandTo3DMatrix(md.ICoordSystem);
   IMatrix invCS = expandTo3DMatrix(md.IInvCS);
   IVector Ix = expandTo3DVector(md.Ix);

   CnCoeff DP(3,1); // (x,y,a), C^1
   IVector image = computePoincareMap(PM4D,Ix + CS*center,invCS,box,2*md.bifNumber,DP); // we compute P, not P^2
   image = changeCoordSystem(Ix,invCS,CS,image,DP);

// next we compute partial derivative dx/da 

   interval det = (DP(0,0)-1)*(DP(1,1)-1)-DP(0,1)*DP(1,0);
   interval dxda = ( DP(0,2)*(1-DP(1,1)) + DP(0,1)*DP(1,2) )/det;

   return dxda;
}

// ----------------------------------------------------------------------------
// this is rigorous routine
// it computes d^2 G/dxda

interval computeD2G_DXDA(PoincareMap& PM4D, const interval& x, const interval& y, const interval& a, const MainData& md)
{
// here we need C^2 integration in four dimensional space
   IVector u(3);
   u[0]=x; u[1]=y; u[2]=a;
   IVector center = midVector(u);
   IVector box = u-center;

   IMatrix CS = expandTo3DMatrix(md.ICoordSystem);
   IMatrix invCS = expandTo3DMatrix(md.IInvCS);
   IVector Ix = expandTo3DVector(md.Ix);

   CnCoeff DP(3,2); // (x,y,a), C^2
   IVector image = computePoincareMap(PM4D,Ix + CS*center,invCS,box,4*md.bifNumber,DP);
   image = changeCoordSystem(Ix,invCS,CS,image,DP);

   interval temp = (1.-DP(1,1));
   interval dydx = DP(1,0)/temp;
   interval dyda = DP(1,2)/temp;
   interval d2ydxda = (dyda*DP(1,0,1)+dydx*(dyda*DP(1,1,1)+DP(1,1,2))+DP(1,0,2))/temp;

   return -d2ydxda*DP(0,1) - dyda*(dydx*DP(0,1,1) + DP(0,0,1)) - dydx*DP(0,1,2) - DP(0,0,2);
}

// ----------------------------------------------------------------------------
// this is rigorous routine
// it computes dG/da

interval computeDGDA(PoincareMap& PM4D, const interval& x, const interval& y, const interval& a, const MainData& md)
{
// here we need C^2 integration in four dimensional space
   IVector u(3);
   u[0]=x; u[1]=y; u[2]=a;

   IVector center = midVector(u);
   IVector box = u-center;

   IMatrix CS = expandTo3DMatrix(md.ICoordSystem);
   IMatrix invCS = expandTo3DMatrix(md.IInvCS);
   IVector Ix = expandTo3DVector(md.Ix);

   CnCoeff DP(3,1); // (x,y,a), C^1
   IVector image = computePoincareMap(PM4D,Ix + CS*center,invCS,box,4*md.bifNumber,DP);
   image = changeCoordSystem(Ix,invCS,CS,image,DP);

   interval dyda = DP(1,2)/(1.-DP(1,1));
   return -DP(0,2) - DP(0,1)*dyda;
}

// ----------------------------------------------------------------------------

// this function compute eigenvalues of the derivative of PoincareMap at the fixed point
template<typename PoincareMapType, typename MainDataType>
typename PoincareMapType::VectorType computeEigenvalues(
      BasicPoincareMap& BPM, 
      PoincareMapType& PM, 
      const MainDataType& md, 
      typename PoincareMapType::ScalarType& o_ev1, 
      typename PoincareMapType::ScalarType& o_ev2, 
      bool findBetterEnclosure
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

   LVector box(2), x(2), N(2);
   DVector z(2);
   DVector sP=md.startPoint;

   typename LInterval::BoundType size1 = diam(md.box[0]).rightBound()/10;
   typename LInterval::BoundType size2 = diam(md.box[1]).rightBound()/10;

   box[0] = LInterval(-1.,1.)*size1;
   box[1] = LInterval(-1.,1.)*size2;

   z = findPeriodicOrbit(BPM,sP,2*md.bifNumber);
   x[0] = z[1];
   x[1] = z[2];

   // now, we verify if x+md.ICoordSystem*box \subset md.Ix + md.ICoordSystem *md.box
   if(!subset(md.IInvCS*(x-md.Ix)+box,md.box))
      throw std::runtime_error("checkEigenvalues: inclusion not satisfied");

   N = provePeriodicOrbit(PM,x,md.ICoordSystem,md.IInvCS,box,2*md.bifNumber);
   if(findBetterEnclosure)
   {
      LVector center = midVector(N);
      try{
            LVector NY = provePeriodicOrbit(PM,x+md.ICoordSystem*center,md.ICoordSystem,md.IInvCS,N-center,2*md.bifNumber);
            x += md.ICoordSystem*center;
            N = NY;
      }catch(std::exception&){std::cout << "Cannot find better enclosure" << std::endl << N << std::endl;}
   }

   LCnCoeff DF(2,1);
   computePoincareMap(PM,x,md.ICoordSystem,N,2*md.bifNumber,DF);
   eigenvalues(LMatrix(DF),o_ev1,o_ev2);
   std::cout << "ev1: " << o_ev1 << std::endl;
   std::cout << "ev2: " << o_ev2 << std::endl << std::endl;
   return md.IInvCS*(x-md.Ix)+N;
}

// ----------------------------------------------------------------------------

// this function compute eigenvalues of the derivative of PoincareMap at the fixed point
// moreover, it computes derivative of eigenvalue as a function of parameter

interval computeEigenvalues(
   PoincareMap& PM, 
   PoincareMap& PM4D, 
   const MainData& md, 
   const interval& a, 
   interval& o_ev1, 
   interval& o_ev2,
   IMatrix&  o_eigenvectors
   )
{
   PM.setParameter("a",a);
   IVector N = provePeriodicOrbit(PM,md.Ix,md.ICoordSystem,md.IInvCS,md.box,2*md.bifNumber);
   // after this procedure we now that in N there is unique fixed point

// next we use C^2 integration in four dimensional space
   IVector u(3);
   u[0]=N[0]; u[1]=N[1]; u[2]=a;
   IVector center = midVector(u);
   IVector box2 = u-center;

   IMatrix CS = expandTo3DMatrix(md.ICoordSystem);
   IMatrix invCS = expandTo3DMatrix(md.IInvCS);
   IVector Ix = expandTo3DVector(md.Ix);

   CnCoeff DP(3,2); // (x,y,a), C^2
   IVector image = computePoincareMap(PM4D,Ix + CS*center,invCS,box2,2*md.bifNumber,DP);
   o_eigenvectors = eigenvectors(IMatrix(DP)); // eigenvectors are computed in standard coordinates
   image = changeCoordSystem(Ix,invCS,CS,image,DP);

   interval dev1;
   eigenvalues(DP,o_ev1,o_ev2,dev1);
   return dev1;
}
